// $Id: CalendarEvent.cpp,v 1.13 2013/11/13 14:34:27 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CalendarEvent.h"
#include "CalendarXmlDef.h"
#include "Str.h"
#include "XmlTimeStamp.h"
#include "XmlString.h"

#include <QTextCharFormat>

namespace CALENDAR
{

    //_________________________________________________________
    CalendarEvent::SelectionMap& CalendarEvent::selectionMap( void )
    {
        static SelectionMap map;
        return map;
    }

    //_________________________________________________________
    CalendarEvent::RepeatMap& CalendarEvent::repeatMap( void )
    {
        static RepeatMap map;
        return map;
    }

    //_________________________________________________________
    CalendarEvent::CalendarEvent( const QDomElement& element ):
        Counter( "CalendarEvent" ),
        flags_( FullDays|Notify ),
        selectionMode_( SelectionRange ),
        repeatMode_( RepeatNone )
    {
        Debug::Throw( "LogEntry::LogEntry [dom].\n" );
        _initializeMaps();

        // parse children elements
        for(QDomNode child_node = element.firstChild(); !child_node.isNull(); child_node = child_node.nextSibling() )
        {
            QDomElement childElement = child_node.toElement();
            if( childElement.isNull() ) continue;
            else if( childElement.tagName() == XML::DESCRIPTION ) setDescription( XmlString( childElement.text() ).toText() );
            else if( childElement.tagName() == XML::REPEAT_MODE ) setRepeatMode( (RepeatMode)childElement.text().toInt() );
            else if( childElement.tagName() == XML::SELECTION_MODE ) setSelectionMode( (SelectionMode)childElement.text().toInt() );
            else if( childElement.tagName() == XML::BEGIN ) begin_.setTime_t( XmlTimeStamp( childElement ) );
            else if( childElement.tagName() == XML::END ) end_.setTime_t( XmlTimeStamp( childElement ) );
            else if( childElement.tagName() == XML::FLAGS ) setFlags( (Flags)childElement.text().toInt() );

            // kept for backward compatibility
            else if( childElement.tagName() == XML::VALIDITY ) setFlag( Valid, (bool)childElement.text().toInt() );
            else if( childElement.tagName() == XML::MESSAGE ) setDescription( XmlString( childElement.text() ).toText() );
        }

    }

    //______________________________________________________________________
    QDomElement CalendarEvent::domElement( QDomDocument& document ) const
    {
        Debug::Throw( "CalendarEvent::domElement.\n" );

        QDomElement out( document.createElement( XML::TAG ) );

        out.
            appendChild( document.createElement( XML::REPEAT_MODE ) ).
            appendChild( document.createTextNode( QString().setNum( repeatMode_ ) ) );

        out.
            appendChild( document.createElement( XML::SELECTION_MODE ) ).
            appendChild( document.createTextNode( QString().setNum( selectionMode_ ) ) );

        out.
            appendChild( document.createElement( XML::FLAGS ) ).
            appendChild( document.createTextNode( QString().setNum( flags_ ) ) );

        // begin and end date
        if( begin_.isValid() ) out.appendChild( XmlTimeStamp( begin_.toTime_t() ).domElement( XML::BEGIN, document ) );
        if( end_.isValid() ) out.appendChild( XmlTimeStamp( end_.toTime_t() ).domElement( XML::END, document ) );

        // message
        if( !description_.isEmpty() )
        {
            out.
                appendChild( document.createElement( XML::MESSAGE ) ).
                appendChild( document.createTextNode( XmlString( description_ ).toXml() ) );
        }

        return out;
    }

    //______________________________________________
    QString CalendarEvent::toString( void ) const
    {

        QString out;
        if( hasFlag( FullDays ) )
        {

            if( begin_.date() == end_.date() ) out += begin_.date().toString() + "\n";
            else {

                out += QString( QObject::tr( "From %1\n" ) ).arg( begin_.date().toString( Qt::TextDate ) );
                out += QString( QObject::tr( "To %1\n" ) ).arg( end_.date().toString( Qt::TextDate ) );

            }

        } else if( begin() == end() ) {

            out += QString( "%1 %2\n" ).arg( begin_.date().toString( Qt::TextDate ) ).arg( begin_.time().toString( "h:mm ap" ) );

        } else if( begin_.date() == end_.date() ) {

            out += begin_.date().toString() + "\n";
            out += QString( QObject::tr( "From %2\n" ) ).arg( begin_.time().toString( "h:mm ap" ) );
            out += QString( QObject::tr( "To %2\n" ) ).arg( end_.time().toString( "h:mm ap" ) );

        } else {

            out += QString( QObject::tr( "From %1 %2\n" ) ).arg( begin_.date().toString( Qt::TextDate ) ).arg( begin_.time().toString( "h:mm ap" ) );
            out += QString( QObject::tr( "To %1 %2\n" ) ).arg( end_.date().toString( Qt::TextDate ) ).arg( end_.time().toString( "h:mm ap" ) );

        }

        // contents
        out += description_;

        return out;

    }

    //______________________________________________
    QString CalendarEvent::toHtml( void ) const
    {

        QString out;
        if( hasFlag( FullDays ) )
        {

            if( begin_.date() == end_.date() )
            {

                out += QString( "<div align='center'><b>%1</b></div>\n" ).arg( begin_.date().toString() );

            } else {

                out += QString( "<div align='center'>\n" );
                out += QString( "<table>\n" );
                out += QString( QObject::tr( "<tr><td align='right'>From</td><td><b> %1</b></td></tr>\n" ) ).arg( begin_.date().toString( Qt::TextDate ) );
                out += QString( QObject::tr( "<tr><td align='right'>To</td><td><b> %1</b></td></tr>\n" ) ).arg( end_.date().toString( Qt::TextDate ) );
                out += QString( "</table>\n" );
                out += QString( "</div>\n" );

            }

        } else if( begin() == end() ) {

            out += QString( "<div align='center'><b>%1 %2</b></div>" ).arg( begin_.date().toString( Qt::TextDate ) ).arg( begin_.time().toString( "h:mm ap" ) );

        } else if( begin_.date() == end_.date() ) {

            out += QString( "<div align='center'><b>%1</b></div>\n" ).arg( begin_.date().toString() );
            out += QString( "<div align='center'>\n" );
            out += QString( "<table>\n" );
            out += QString( QObject::tr( "<tr><td align='right'>From</td><td><b> %2</b></td></tr>\n" ) ).arg( begin_.time().toString( "h:mm ap" ) );
            out += QString( QObject::tr( "<tr><td align='right'>To</td><td><b> %2</b></td></tr>\n" ) ).arg( end_.time().toString( "h:mm ap" ) );
            out += QString( "</table>\n" );
            out += QString( "</div>\n" );

        } else {

            out += QString( "<div align='center'>\n" );
            out += QString( "<table>\n" );
            out += QString( QObject::tr( "<tr><td align='right'>From</td><td><b> %1 %2</b></td></tr>\n" ) ).arg( begin_.date().toString( Qt::TextDate ) ).arg( begin_.time().toString( "h:mm ap" ) );
            out += QString( QObject::tr( "<tr><td align='right'>To</td><td><b> %1 %2</b></td></tr>\n" ) ).arg( end_.date().toString( Qt::TextDate ) ).arg( end_.time().toString( "h:mm ap" ) );
            out += QString( "</table>\n" );
            out += QString( "</div>\n" );
        }

        // contents
        out += QString( "<div align='center'>%1</div>\n" ).arg( QString( description() ).replace( "\n", "<br/>" ) );

        return out;

    }

    //______________________________________________
    void CalendarEvent::toLayout( QTextCursor& cursor ) const
    {

        QTextCharFormat defaultFormat( cursor.charFormat() );
        QTextCharFormat boldFormat( defaultFormat );
        boldFormat.setFontWeight( QFont::Bold );

        if( hasFlag( FullDays ) )
        {
            if( begin_.date() == end_.date() )
            {

                cursor.insertText( begin_.date().toString() + "\n", boldFormat );

            } else {

                cursor.insertText( QObject::tr( "From " ), defaultFormat );
                cursor.insertText( begin_.date().toString() + "\n", boldFormat );
                cursor.insertText( QObject::tr( "To " ), defaultFormat );
                cursor.insertText( end_.date().toString() + "\n", boldFormat );

            }

        } else if( begin_ == end_ ) {

            cursor.insertText( begin_.date().toString( Qt::TextDate ) + " " + begin_.time().toString( "h:mm ap" ) + "\n", boldFormat );

        } else {

            cursor.insertText( QObject::tr( "From " ), defaultFormat );
            cursor.insertText( begin_.date().toString( Qt::TextDate ) + " " + begin_.time().toString( "h:mm ap" ) + "\n", boldFormat );
            cursor.insertText( QObject::tr( "To " ), defaultFormat );
            cursor.insertText( end_.date().toString( Qt::TextDate ) + " " + end_.time().toString( "h:mm ap" ) + "\n", boldFormat );

        }

        cursor.insertText( description_, defaultFormat );

        switch( repeatMode_ )
        {

            default:
            case RepeatNone:
            break;

            case RepeatWeek:
            cursor.insertText( QObject::tr( "\nRepeat every week" ) );
            break;

            case RepeatMonth:
            cursor.insertText( QObject::tr( "\nRepeat every month" ) );
            break;

            case RepeatYear:
            cursor.insertText( QObject::tr( "\nRepeat every year" ) );
            break;

        }

    }

    //______________________________________________
    void CalendarEvent::generate( void )
    {
        ranges_.clear();

        // check validity
        if( !(flags_&Valid) ) return;

        // check selection mode
        if( selectionMode() == SelectionRange ) ranges_ << Range( begin_.date(), end_.date() );
        else {

            int firstDay = begin_.date().dayOfWeek();
            int lastDay = end_.date().dayOfWeek();
            if( lastDay < firstDay ) std::swap( firstDay, lastDay );

            int firstWeek = begin_.date().weekNumber();
            int lastWeek = end_.date().weekNumber();
            if( lastWeek < firstWeek ) std::swap( firstWeek, lastWeek );

            QDate firstDate = begin_.date();
            if( firstDate.dayOfWeek() != firstDay ) firstDate = firstDate.addDays( firstDay - lastDay );
            for( int week = 0; week <= lastWeek-firstWeek; week++ )
            {
                // this does not work
                ranges_ << Range(
                    firstDate.addDays( 7*week ),
                    firstDate.addDays( 7*week + lastDay - firstDay ) );
            }

        }

        return;

    }

    //______________________________________________________________________
    void CalendarEvent::generate( const QDate& first, const QDate& second )
    {

        generate();

        RangeList local( ranges() );
        ranges_.clear();

        foreach( const Range& range, local )
        {
            if( repeatMode() == RepeatNone )
            {

                // no repeat mode
                if( OverlapsFTor( first, second )( range ) ) ranges_ << range;

            } else if( repeatMode() == RepeatWeek ) {

                int firstDay = range.first.dayOfWeek();
                int secondDay = range.second.dayOfWeek();
                for( int day = first.day(); day <= second.day(); day++ )
                {

                    QDate current( first.year(), first.month(), day );
                    int currentDay( current.dayOfWeek() );
                    if(
                        ( firstDay <= secondDay && currentDay >= firstDay && currentDay <= secondDay ) ||
                        ( firstDay > secondDay && (currentDay <= secondDay || currentDay >= firstDay ) ) )
                    { ranges_ << Range( current, current ); }

                }


            } else if( repeatMode() == RepeatMonth ) {


                if( range.first.year() == range.second.year() && range.first.month() == range.second.month() )
                {

                    ranges_ << Range(
                        QDate( first.year(), first.month(), range.first.day() ),
                        QDate( first.year(), first.month(), qMin( range.second.day(), second.day() ) ) );

                } else {

                    ranges_ << Range( first, QDate( first.year(), first.month(), range.second.day() ) );
                    if( range.first.day() <= second.day() )
                    { ranges_ << Range( QDate( first.year(), first.month(), range.first.day() ), second ); }

                }

            } else if( repeatMode() == RepeatYear ) {

                if( range.first.year() == range.second.year() )
                {
                    if(
                        ( range.first.month() < first.month() || (range.first.month() == first.month() && range.first.day() <= second.day() ) ) &&
                        range.second.month() >= second.month() )
                    {

                        ranges_ << Range(
                            qMax( first, QDate( first.year(), range.first.month(), range.first.day() ) ),
                            qMin( second, QDate( second.year(), range.second.month(), range.second.day() ) ) );

                    }

                } else {

                    if( QDate( first.year(), range.second.month(), range.second.day() ) >= first )
                    { ranges_ << Range( first, qMin(  QDate( first.year(), range.second.month(), range.second.day() ), second ) ); }

                    if( QDate( second.year(), range.first.month(), range.first.day() ) <= second )
                    { ranges_ << Range( qMax( first, QDate( second.year(), range.first.month(), range.first.day() ) ), second ); }

                }

            }

        }

        return;

    }

    //_______________________________________________________________
    void CalendarEvent::_initializeMaps( void )
    {

        if( selectionMap().empty() )
        {
            selectionMap().insert( CalendarEvent::SelectionRange, QObject::tr( "Range" ) );
            selectionMap().insert( CalendarEvent::SelectionBox, QObject::tr( "Box" ) );
        }

        if( repeatMap().empty() )
        {
            repeatMap().insert( CalendarEvent::RepeatNone, QObject::tr( "Once" ) );
            repeatMap().insert( CalendarEvent::RepeatWeek, QObject::tr( "Every Week" ) );
            repeatMap().insert( CalendarEvent::RepeatMonth, QObject::tr( "Every Month" ) );
            repeatMap().insert( CalendarEvent::RepeatYear, QObject::tr( "Every Year" ) );
        }

    }


    //__________________________________________________
    QString CalendarEvent::validate( void )
    {

        setFlag( Valid, false );
        if( begin_ > end_ ) return QObject::tr( "Tag begin date is later than its end date." );

        if( repeatMode() == CalendarEvent::RepeatWeek && begin_.daysTo( end_ ) >= 6 )
        {
            return
                QObject::tr( "Tag must be repeated every week but it covers a period larger than a week.\n"
                "As a result, the entire calendar would be selected." );
        }

        if( repeatMode() == CalendarEvent::RepeatMonth && begin_.date().month() != end_.date().month() && begin_.date().day() < end_.date().day() )
        {
            return
                QObject::tr( "Tag must be repeated every month but it covers a period larger than a month.\n"
                "As a result, the entire calendar would be selected." );
        }

        if( repeatMode() == CalendarEvent::RepeatYear && begin_.date().year() != end_.date().year() && begin_.date() < QDate( begin_.date().year(), end_.date().month(), end_.date().day() ) )
        {
            return
                QObject::tr( "Tag must be repeated yearly but it covers a pereiod larger than a year.\n"
                "As a result, the entire calendar would be selected." );
        }

        setFlag( Valid, true );
        return QString();
    }
}
