#ifndef _CalendarEvent_h_
#define _CalendarEvent_h_
// $Id: CalendarEvent.h,v 1.8 2013/04/29 18:27:15 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "Debug.h"

#include <QDate>
#include <QDomDocument>
#include <QDomElement>
#include <QString>
#include <QList>
#include <QMap>

#include <QTextCursor>

#include <algorithm>

//! Calendar widgets
namespace CALENDAR
{

    //! transparent Calendar widget
    class CalendarEvent: public Counter
    {

        public:

        typedef QList<CalendarEvent> List;

        //! constructor
        CalendarEvent( void ):
            Counter( "CalendarEvent" ),
            flags_( FullDays|Notify ),
            selectionMode_( SelectionRange ),
            repeatMode_( RepeatNone )
        { _initializeMaps(); }

        //! constructor from DOM element
        CalendarEvent( const QDomElement& element );

        //! make dom element
        QDomElement domElement( QDomDocument& document ) const;

        //! equal to operator
        bool operator == (const CalendarEvent& other ) const
        { return begin() == other.begin() && end() == other.end(); }

        //! less than operator
        bool operator < (const CalendarEvent& other ) const
        { return begin() < other.begin() || ( begin() != other.begin() && end() < other.end() ); }

        //!@name accessors
        //@{

        //! begin
        const QDateTime& begin( void ) const
        { return begin_; }

        //! end
        const QDateTime& end( void ) const
        { return end_; }

        //! flags
        enum Flag
        {
            FullDays = 1<<0,
            Notify = 1<<1,
            Notified = 1<<2,
            Valid = 1<<3
        };

        Q_DECLARE_FLAGS( Flags, Flag );

        //! flags
        Flags flags( void ) const
        { return flags_; }

        //! flags
        bool hasFlag( Flag flag ) const
        { return flags_&flag; }

        //! selection mode
        enum SelectionMode
        {
            SelectionRange,
            SelectionBox
        };

        //! selection mode
        const SelectionMode& selectionMode( void ) const
        { return selectionMode_; }

        //! selection name
        static QString SelectionName( SelectionMode& mode )
        { return selectionMap()[mode]; }

        //! selection name
        QString selectionName( void ) const
        { return selectionMap()[selectionMode()]; }

        //! selection mode
        typedef QMap< unsigned int, QString > SelectionMap;
        static SelectionMode selectionMode( const QString& name )
        {
            for( SelectionMap::const_iterator iter = selectionMap().begin(); iter != selectionMap().end(); iter++ )
            { if( iter.value() == name ) return (SelectionMode) iter.key(); }

            return SelectionRange;
        }

        //! selection map
        static SelectionMap& selectionMap( void );

        //! repeat mode
        enum RepeatMode
        {
            RepeatNone,
            RepeatWeek,
            RepeatMonth,
            RepeatYear
        };

        //! repeat mode
        const RepeatMode& repeatMode( void ) const
        { return repeatMode_; }

        //! repeat name
        static QString repeatModeName( const RepeatMode& mode )
        { return repeatMap()[mode]; }

        //! repeat name
        QString repeatModeName( void ) const
        { return repeatModeName( repeatMode() ); }

        //! repeat map
        typedef QMap< unsigned int, QString > RepeatMap;
        static RepeatMode repeatMode( const QString& name )
        {
            for( RepeatMap::const_iterator iter = repeatMap().begin(); iter != repeatMap().end(); iter++ )
            { if( iter.value() == name ) return (RepeatMode) iter.key(); }

            return RepeatNone;
        }

        //! selection map
        static RepeatMap& repeatMap( void );

        //! description
        const QString& description( void ) const
        { return description_; }

        //! ranges
        typedef QPair<QDate, QDate> Range;
        typedef QList<Range> RangeList;

        //! ranges
        const RangeList& ranges( void ) const
        { return ranges_; }

        //! convert to string
        QString toString( void ) const;

        //! convert to html
        QString toHtml( void ) const;

        //! convert to text document
        void toLayout( QTextCursor& ) const;

        //@}

        //@name modifiers
        //@{

        //! begin
        void setBegin( const QDate& value )
        { begin_ = QDateTime( value, QTime( 0, 0, 0 ) ); }

        //! begin
        void setBegin( const QDateTime& value )
        { begin_ = value; }

        //! end
        void setEnd( const QDate& value )
        { end_ = QDateTime( value, QTime( 23, 59, 59 ) ); }

        //! end
        void setEnd( const QDateTime& value )
        { end_ = value; }

        //! range
        void setRange( const QDate& begin, const QDate& end )
        { setBegin( begin ); setEnd( end ); }

        //! range
        void setRange( const QDateTime& begin, const QDateTime& end )
        { setBegin( begin ); setEnd( end ); }

        //! range
        void setRange( const QDate& begin )
        { setBegin( begin ); setEnd( begin ); }

        //! range
        void setRange( const QDateTime& begin )
        { setBegin( begin ); setEnd( begin ); }

        //! set flags
        void setFlags( Flags flags )
        { flags_ = flags; }

        //! flags
        void setFlag( Flag flag, bool value )
        {
            if( value ) flags_ |= flag;
            else flags_ &= ~flag;
        }

        //! selection mode
        void setSelectionMode( const SelectionMode& value )
        { selectionMode_ = value; }

        //! repeat mode
        void setRepeatMode( const RepeatMode& value )
        { repeatMode_ = value; }

        //! description
        void setDescription( const QString& value )
        { description_ = value; }

        //! generate default range
        void generate( void );

        //! generate list of ranges matching a given date
        /*! this is necessary to easily handle repeat modes */
        void generate( const QDate&, const QDate& );

        //! validate. Returns error message if invalid
        QString validate( void );

        //@}

        //! used to select tags that contains given date
        class ContainsFTor
        {

            public:

            //! constructor
            ContainsFTor( const QDate& date ):
                date_( date )
            {}

            //! prediction
            bool operator() ( const CalendarEvent& tag ) const
            { return std::find_if( tag.ranges_.begin(), tag.ranges_.end(), (*this) ) != tag.ranges_.end(); }

            //! prediction
            bool operator() ( const CalendarEvent::Range& range ) const
            { return date_ >= range.first && date_ <= range.second; }

            private:

            //! predicted date
            QDate date_;

        };

        //! used to select tags that contains given date
        class OverlapsFTor
        {

            public:

            //! constructor
            OverlapsFTor( const QDate& first, const QDate& second ):
                first_( first ),
                second_( second )
            {}

            //! prediction
            bool operator() ( const CalendarEvent& tag ) const
            { return std::find_if( tag.ranges_.begin(), tag.ranges_.end(), (*this) ) != tag.ranges_.end(); }

            //! prediction
            bool operator() ( const CalendarEvent::Range& range ) const
            { return first_ <= range.second && second_ >= range.first; }

            private:

            //! predicted date
            QDate first_;

            //! predicted date
            QDate second_;

        };

        //! returns true if tag contains date
        bool contains( const QDate& date ) const
        { return ContainsFTor( date )( *this ); }

        private:

        //! begin date
        QDateTime begin_;

        //! end date
        QDateTime end_;

        //! flags
        Flags flags_;

        //! mode
        SelectionMode selectionMode_;

        //! mode
        RepeatMode repeatMode_;

        //! description
        QString description_;

        //! ranges that match a given date
        RangeList ranges_;

        //! init maps
        static void _initializeMaps( void );

    };

}

Q_DECLARE_OPERATORS_FOR_FLAGS( CALENDAR::CalendarEvent::Flags );

#endif
