
// $Id: CalendarEventList.cpp,v 1.8 2013/04/29 18:27:15 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Debug.h"
#include "CalendarEventList.h"
#include "CalendarXmlDef.h"
#include "XmlDocument.h"

#include <algorithm>

namespace CALENDAR
{

    //____________________________________________________________________________
    bool CalendarEventList::read( File file )
    {

        Debug::Throw( "CalendarEventList::read.\n" );

        // clear list
        events_.clear();

        if( !file.isEmpty() ) file_ = file;
        if( file_.isEmpty() ) return false;

        // parse the file
        XmlDocument document;
        QFile qfile( file );
        if ( !document.setContent( &qfile ) )
        {
            Debug::Throw() << document.error() << endl;
            return false;
        }

        return _read( document, events_ );

    }

    //____________________________________________________________________________
    bool CalendarEventList::write( File file ) const
    {

        if( file.isEmpty() ) file = file_;
        if( file.isEmpty() ) return false;

        // create document and read
        XmlDocument document;
        {
            QFile qtfile( file );
            document.setContent( &qtfile );
        }

        // read old values
        CalendarEvent::List events;
        if( _read( document, events ) && !_differs( events_, events ) ) return true;

        // create main element
        QDomElement top = document.createElement( XML::TAG_LIST );
        foreach( const CalendarEvent& event, events_ )
        { if( event.hasFlag( CalendarEvent::Valid ) )  top.appendChild( event.domElement( document ) ); }

        // append top node to document and write
        document.replaceChild( top );
        {
            QFile qfile( file );
            if( !qfile.open( QIODevice::WriteOnly ) ) return false;
            qfile.write( document.toByteArray() );
        }

        return true;
    }

    //____________________________________________________________________________
    void CalendarEventList::add( const CalendarEvent& event )
    { events_.append( event ); }

    //____________________________________________________________________________
    void CalendarEventList::remove( const CalendarEvent& event )
    {
        if( std::find( events_.begin(), events_.end(), event ) != events_.end() )
        { events_.erase( std::remove( std::find( events_.begin(), events_.end(), event ), events_.end(), event ) ); }
    }

    //____________________________________________________________________________
    bool CalendarEventList::_read( const XmlDocument& document, CalendarEvent::List& events ) const
    {

        // look for relevant element
        QDomNodeList topNodes = document.elementsByTagName( XML::TAG_LIST );
        if( topNodes.isEmpty() ) return false;
        for(QDomNode node = topNodes.at(0).firstChild(); !node.isNull(); node = node.nextSibling() )
        {

            QDomElement element = node.toElement();
            if( element.isNull() ) continue;
            if( element.tagName() == XML::TAG ) events.append( element );

        }

        return true;
    }

    //____________________________________________________________________________
    bool CalendarEventList::_differs( const CalendarEvent::List& first, const CalendarEvent::List& second ) const
    {

        // check sizes
        if( first.size() != second.size() ) return true;

        // check records one by one
        for( int i = 0; i < first.size(); ++i )
        {

            // records
            const CalendarEvent& firstEvent = first[i];
            const CalendarEvent& secondEvent = second[i];

            if( !( firstEvent == secondEvent ) ) return true;
            else if( firstEvent.flags() != secondEvent.flags() ) return true;
            else if( firstEvent.description() != secondEvent.description() ) return true;
            else if( firstEvent.selectionMode() != secondEvent.selectionMode() ) return true;
            else if( firstEvent.repeatMode() != secondEvent.repeatMode() ) return true;

        }

        return false;

    }

}
