// $Id: Alarm.cpp,v 1.13 2013/04/29 18:27:15 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

/*!
\file    Alarm.cpp
\brief   Alarm settings
\author  Hugo Pereira
\version $Revision: 1.13 $
\date    $Date: 2013/04/29 18:27:15 $
*/

#include "Alarm.h"
#include "ClockXmlDef.h"
#include "Str.h"
#include "XmlTimeStamp.h"
#include "XmlString.h"

namespace CLOCK
{

    //__________________________________________________________________________
    Alarm::Alarm( const QDomElement& element ):
        Counter( "Alarm" ),
        days_( (1<<8)-1 ),
        time_( TimeStamp::now() ),
        enabled_( false )
    {
        Debug::Throw( "Alarm::Alarm [dom].\n" );

        // parse attributes
        QDomNamedNodeMap attributes( element.attributes() );
        for( unsigned int i=0; i<attributes.length(); i++ )
        {
            QDomAttr attribute( attributes.item( i ).toAttr() );
            if( attribute.isNull() ) continue;
            if( attribute.name() == XML::ENABLED ) setEnabled( attribute.value().toInt() );
            else if( attribute.name() == XML::DAYS ) setDays( (unsigned int) attribute.value().toInt() );
            else Debug::Throw(0) << "Alarm::Alarm - unrecognized entry attribute: \"" << attribute.name() << "\"\n";
        }

        // parse children elements
        for(QDomNode child_node = element.firstChild(); !child_node.isNull(); child_node = child_node.nextSibling() )
        {
            QDomElement child_element = child_node.toElement();
            if( child_element.isNull() ) continue;
            if( child_element.tagName() == XML::TEXT ) setText( XmlString( child_element.text() ).toText() );
            else if( child_element.tagName() == XML::TIME ) setTime( XmlTimeStamp( child_element ) );
        }

    }

    //______________________________________________________________________
    QDomElement Alarm::domElement( QDomDocument& document ) const
    {
        Debug::Throw( "Alarm::domElement.\n" );

        QDomElement out( document.createElement( XML::ALARM ) );
        out.setAttribute( XML::ENABLED, Str().assign<bool>( enabled() ) );
        out.setAttribute( XML::DAYS, Str().assign<unsigned int>( days() ) );

        if( time().isValid() ) out.appendChild( XmlTimeStamp( time() ).domElement( XML::TIME, document ) );
        if( !text().isEmpty() )
        {
            out.
                appendChild( document.createElement( XML::TEXT ) ).
                appendChild( document.createTextNode( XmlString( text() ).toXml() ) );
        }

        return out;
    }

    //______________________________________________________________________
    QString Alarm::dayNames( unsigned int days )
    {
        QString out;
        QTextStream outStr( &out );
        bool first( true );
        for( unsigned int i = 1; i < 8; ++i )
        {
            if( !(days&(1<<i)) ) continue;

            if( first ) first = false;
            else outStr << ", ";
            outStr << QDate::shortDayName( i );

        }
        return out;
    }


    //______________________________________________________________________
    bool Alarm::triggered( const QDate& date, const QTime& time )
    {
        if( !enabled_ ) return false;

        // check for "direct" triggering
        if(
            date != last_trigger_date_ &&
            time.hour() == time_.hours() &&
            time.minute() == time_.minutes() &&
            hasDay( date.dayOfWeek() ) )
        {
            last_trigger_date_ = date;
            return true;
        }

        // check for "repeat" triggering
        if(
            repeat_ &&
            repeatInterval_ > 0 &&
            !( QTime( time_.hours(), time_.minutes() ).secsTo( time ) % repeatInterval_ ) )
        {
            repeat_ = false;
            return true;
        }

        return false;
    }

}
