// $Id: ClockWidget.cpp,v 1.64 2013/11/15 14:50:40 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "AdjustTimeDialog.h"
#include "ClockConfigurationDialog.h"

#include "ClockWidget.h"
#include "ClockWidget.moc"

#include "AlarmDialog.h"
#include "AlarmListDialog.h"
#include "BaseContextMenu.h"
#include "Color.h"
#include "Debug.h"
#include "Effects.h"
#include "Icons.h"
#include "IconEngine.h"
#include "InformationDialog.h"
#include "QtUtil.h"
#include "RoundedPath.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QDateTime>
#include <QProcess>

#include <QApplication>
#include <QPainter>

namespace CLOCK
{

    //_______________________________________________
    ClockWidget::ClockWidget( QWidget* parent ):
        MainWindow( parent ),
        delay_( 0 ),
        alarms_( this )
    {

        Debug::Throw( "CLOCK::ClockWidget::ClockWidget.\n" );
        setWindowTitle( Util::windowTitle( tr( "Transparent Clock" ) ) );

        // actions
        _installActions();
        _configurationAction().setText( tr( "Transparent Clock Settings..." ) );

        //! update configuration
        _updateConfiguration();

        //! read alarms
        alarms_.read( QString( XmlOptions::get().raw( "RC_FILE" ) ) );
        if( !alarms_.alarms().empty() ) alarmTimer_.start( 1000, this );

        // start timer. update every second
        timer_.start( 1000, this );

        // write alarms at exit
        connect( qApp, SIGNAL(aboutToQuit()), &alarms_, SLOT(write()) );

    }

    //______________________________________________________________________
    void ClockWidget::_installContextMenuActions( BaseContextMenu* menu )
    {
        Debug::Throw( "CLOCK::ClockWidget::_installContextMenuActions.\n" );

        // get menu from parent
        MainWindow::_installContextMenuActions( menu );

        // customize
        menu->insertSeparator( &_menuSeparator() );
        menu->insertAction( &_menuSeparator(), showHourTicksAction_ );
        menu->insertAction( &_menuSeparator(), showMinuteTicksAction_ );
        menu->insertAction( &_menuSeparator(), showSecondsAction_ );
        menu->insertAction( &_menuSeparator(), showDateAction_ );
        menu->insertSeparator( &_menuSeparator() );

        menu->insertAction( &_menuSeparator(), configureAlarmsAction_ );
        menu->insertAction( &_menuSeparator(), adjustTimeAction_ );

    }

    //______________________________________________________________________
    void ClockWidget::enterEvent( QEvent* event )
    {
        update();
        MainWindow::enterEvent( event );
    }

    //______________________________________________________________________
    void ClockWidget::timerEvent( QTimerEvent* event )
    {

        if( event->timerId() == timer_.timerId() ) update();
        else if( event->timerId() == alarmTimer_.timerId() ) _checkAlarms();
        else MainWindow::timerEvent( event );
        return;

    }

    //______________________________________________________________________
    void ClockWidget::_installActions( void )
    {

        Debug::Throw( "CLOCK::ClockWidget::_installActions.\n" );

        // alarms
        addAction( configureAlarmsAction_ = new QAction( IconEngine::get( ICONS::ALARM ), tr( "Configure Alarms..." ), this ) );
        configureAlarmsAction_->setToolTip( tr( "Add, remove or modify alarms" ) );
        connect( configureAlarmsAction_, SIGNAL(triggered()), SLOT(_configureAlarms()) );

        addAction( showHourTicksAction_ = new QAction( tr( "Show Ticks (hours)" ), this ) );
        showHourTicksAction_->setCheckable( true );
        showHourTicksAction_->setChecked( false );
        connect( showHourTicksAction_, SIGNAL(toggled(bool)), SLOT(_toggleHourTicks(bool)) );

        addAction( showMinuteTicksAction_ = new QAction( tr( "Show Ticks (minutes)" ), this ) );
        showMinuteTicksAction_->setCheckable( true );
        showMinuteTicksAction_->setChecked( false );
        connect( showMinuteTicksAction_, SIGNAL(toggled(bool)), SLOT(_toggleMinuteTicks(bool)) );

        addAction( showSecondsAction_ = new QAction( tr( "Show Seconds" ), this ) );
        showSecondsAction_->setCheckable( true );
        showSecondsAction_->setChecked( false );
        connect( showSecondsAction_, SIGNAL(toggled(bool)), SLOT(_toggleSeconds(bool)) );

        addAction( showDateAction_ = new QAction( tr( "Show Date" ), this ) );
        showDateAction_->setCheckable( true );
        showDateAction_->setChecked( false );
        connect( showDateAction_, SIGNAL(toggled(bool)), SLOT(_toggleDate(bool)) );

        addAction( adjustTimeAction_ = new QAction( tr( "Adjust Time..." ), this ) );
        adjustTimeAction_->setToolTip( tr( "Run external command to adjust system date and time" ) );
        connect( adjustTimeAction_, SIGNAL(triggered()), SLOT(_adjustTime()) );


    }

    //______________________________________________________________________
    void ClockWidget::_paint( QPaintDevice& device, const QRect& clip )
    {

        // get current time
        QTime time( QTime::currentTime() );
        QDate date( QDate::currentDate() );
        int hour( time.hour() );
        if( delay_ )
        {
            hour += delay_;
            while( hour < 0 ) { hour += 24; date = date.addDays( -1 ); }
            while( hour >= 24 ) { hour -= 24; date = date.addDays( 1 ); }
        }

        QPixmap pixmap( size() );
        pixmap.fill( Qt::transparent );

        {

            // prepare painter
            static const double virtual_side = 1000;
            QPainter painter( &pixmap );
            painter.setClipRect( clip );
            painter.setRenderHints(QPainter::Antialiasing );

            int width( ClockWidget::width() );
            int height( ClockWidget::height() );
            if( _showDateAction().isChecked() )
            {
                int labelHeight( QFontMetrics( font() ).height() );
                height -= ( labelHeight + _margins().top() );

                // create rect for painting text
                QRect rect( ClockWidget::rect() );
                rect.setTop( rect.bottom() - labelHeight - _margins().top() );

                // retrieve current date
                painter.save();
                painter.setPen(  foregroundColor() );

                QString buffer;
                QTextStream( &buffer ) << date.day() << "/" << date.month() << "/" << date.year();
                painter.drawText( rect, Qt::AlignCenter, buffer );
                painter.restore();

            }

            painter.translate(  width/2, height/2  );

            static const int arrow_width( 30 );
            int side = qMin( width-_margins().width(), height - _margins().height() ) ;
            painter.scale(side/virtual_side, side/virtual_side );

            // draw hours ticks
            if( _showHourTicksAction().isChecked() )
            {

                painter.save();
                painter.setPen( Qt::NoPen );
                painter.setBrush( hourColor_ );
                for (int j = 0; j < 12; ++j)
                {
                    if( j%3 == 0 ) painter.drawEllipse( QPointF(485, 0 ), 30, 30 );
                    painter.rotate(30.0);
                }
                painter.restore();

            }

            // draw minutes ticks
            if( _showMinuteTicksAction().isChecked() )
            {

                // draw minutes ticks
                painter.save();
                painter.setPen( minuteColor_ );
                for (int j = 0; j < 60; ++j)
                {
                    if ((j % 5) != 0) painter.drawPoint(500, 0);
                    painter.rotate(6.0);
                }
                painter.restore();

            }

            // draw center
            painter.setPen( Qt::NoPen );
            painter.save();
            painter.setBrush( hourColor_ );

            painter.drawEllipse( QPointF(0,0), arrow_width, arrow_width );

            // draw hours
            painter.rotate(30.0 * ( (hour + time.minute() / 60.0)));

            static const int rounding = 25;
            static const QPolygonF hour_hand( _hand( arrow_width, 400 ) );
            painter.drawPath( RoundedPath( hour_hand, rounding ) );
            painter.restore();

            // draw minutes
            painter.save();
            painter.setBrush( minuteColor_ );
            painter.rotate(6.0 * (time.minute() + time.second() / 60.0));

            static const QPolygonF minute_hand( _hand( arrow_width, 450 ) );
            painter.drawPath( RoundedPath( minute_hand, rounding ) );
            painter.restore();

            // draw seconds
            if( _showSecondsAction().isChecked() )
            {

                painter.save();
                painter.setBrush( secondColor_ );
                painter.rotate( 6.0*time.second() );

                static const QPolygonF second_hand( _hand( arrow_width, 450 ) );
                painter.drawPath( RoundedPath( second_hand, rounding ) );
                painter.restore();

            }

            painter.end();

        }

        if( !_showSecondsAction().isChecked() )
        {

            /*
            if seconds are not paint,
            adjust timer delay so that next refresh
            will be just after next minute change
            */
            timer_.start( 1000*(60-time.second() ) , this );

        }

        QPainter painter( &device );
        painter.setClipRect( clip );
        painter.setRenderHints(QPainter::Antialiasing );

        if( shadowOffset() > 0 && shadowColor().isValid() )
        {
            QImage image( pixmap.toImage() );
            TRANSPARENCY::Effects::shadowBlur(image, shadowOffset(), shadowColor() );
            painter.drawImage(rect().topLeft(), image);
        }

        painter.drawPixmap( rect().topLeft(), pixmap );

        return;

    }

    //________________________________________________
    void ClockWidget::_setRepeat( const Alarm& alarm )
    {

        Debug::Throw( "CLOCK::ClockWidget::_setRepeat.\n" );

        Alarm::List& alarms( alarms_.alarms() );
        for( Alarm::List::iterator iter = alarms.begin(); iter != alarms.end(); ++iter )
        { if( *iter == alarm ) iter->setRepeat( true ); }
    }

    //________________________________________________
    void ClockWidget::_unsetRepeat( const Alarm& alarm )
    {

        Debug::Throw( 0, "CLOCK::ClockWidget::_unsetRepeat.\n" );

        Alarm::List& alarms( alarms_.alarms() );
        for( Alarm::List::iterator iter = alarms.begin(); iter != alarms.end(); ++iter )
        { if( *iter == alarm ) iter->setRepeat( false ); }
    }

    //________________________________________________
    void ClockWidget::_configuration( void )
    {

        Debug::Throw( "CLOCK::ClockWidget::_configuration.\n" );
        ClockConfigurationDialog dialog;
        _configureFromDialog( dialog );

    }

    //________________________________________________
    void ClockWidget::_updateConfiguration( void )
    {
        Debug::Throw( "CLOCK::ClockWidget::_updateConfiguration.\n" );

        // colors
        QColor color;
        bool useThemedColors( !XmlOptions::get().get<bool>( "USE_DEFAULT_COLORS" ) );
        _setHourColor( useThemedColors && (color = XmlOptions::get().get<BASE::Color>( "HOUR_COLOR" ) ).isValid() ? color:foregroundColor() );
        _setMinuteColor( useThemedColors && (color = XmlOptions::get().get<BASE::Color>( "MINUTE_COLOR" ) ).isValid() ? color:foregroundColor() );
        _setSecondColor( useThemedColors && (color = XmlOptions::get().get<BASE::Color>( "SECOND_COLOR" ) ).isValid() ? color:foregroundColor() );

        // intensity
        hourColor_.setAlpha( foregroundIntensity() );
        minuteColor_.setAlpha( foregroundIntensity() );
        secondColor_.setAlpha( foregroundIntensity() );

        // delay
        delay_ = XmlOptions::get().get<int>( "CLOCK_DELAY" );

        // flags
        _showHourTicksAction().setChecked( XmlOptions::get().get<bool>( "SHOW_HOUR_TICKS" ) );
        _showMinuteTicksAction().setChecked( XmlOptions::get().get<bool>( "SHOW_MINUTE_TICKS" ) );
        _showSecondsAction().setChecked( XmlOptions::get().get<bool>( "SHOW_SECONDS" ) );
        _showDateAction().setChecked( XmlOptions::get().get<bool>( "SHOW_DATE" ) );

        // base class update
        MainWindow::_updateConfiguration();
        update();

    }

    //________________________________________________
    void ClockWidget::_toggleHourTicks( bool state )
    {
        XmlOptions::get().set<bool>( "SHOW_HOUR_TICKS", state );
        update();
    }

    //________________________________________________
    void ClockWidget::_toggleMinuteTicks( bool state )
    {
        Debug::Throw() << "CLOCK::ClockWidget::_toggleMinuteTicks - " << state << endl;
        XmlOptions::get().set<bool>( "SHOW_MINUTE_TICKS", state );
        update();
    }

    //________________________________________________
    void ClockWidget::_toggleSeconds( bool state )
    {

        XmlOptions::get().set<bool>( "SHOW_SECONDS", state );
        if( state ) timer_.start( 1000, this );
        update();

    }

    //________________________________________________
    void ClockWidget::_toggleDate( bool state )
    {
        XmlOptions::get().set<bool>( "SHOW_DATE", state );
        update();
    }

    //________________________________________________
    void ClockWidget::_configureAlarms( void )
    {

        Debug::Throw( "CLOCK::ClockWidget::_configureAlarms.\n" );

        // create dialog
        AlarmListDialog dialog( this, alarms_.alarms() );
        if( dialog.centerOnDesktop().exec() == QDialog::Rejected ) return;

        // update
        alarms_.setAlarms( dialog.list() );
        alarms_.write();

        // modify timer accordingly
        if( alarms_.alarms().empty() ) alarmTimer_.stop();
        else alarmTimer_.start( 1000, this );

        return;

    }

    //________________________________________________
    void ClockWidget::_adjustTime( void )
    {

        Debug::Throw( "CLOCK::ClockWidget::_adjustTime.\n" );

        const bool alwaysUseCommand( XmlOptions::get().get<bool>( "ALWAYS_USE_CMD" ) );
        QString command( XmlOptions::get().contains( "ADJUST_TIME_CMD" ) ? QString( XmlOptions::get().raw( "ADJUST_TIME_CMD" ) ).trimmed():"" );

        if( !( alwaysUseCommand && !command.isEmpty() ) )
        {

            // create dialog
            AdjustTimeDialog dialog( this );
            dialog.setCommand( command );
            dialog.setAlwaysUse( alwaysUseCommand );
            if( dialog.exec() == QDialog::Rejected ) return;

            // retrieve command
            command = dialog.command().trimmed();

            // check if valid
            if( command.isEmpty() )
            {
                InformationDialog( this, tr( "No command set to adjust time. Canceled." ) ).exec();
                return;
            }

            XmlOptions::get().set<bool>( "ALWAYS_USE_CMD", dialog.alwaysUse() );

        }

        // try run
        int exitCode( QProcess::execute( command ) );
        if( exitCode != 0 )
        {
            const QString buffer = QString( tr( "Command '%1' excited with code %2" ) ).arg( command ).arg( exitCode );
            InformationDialog( this, buffer ).exec();

            XmlOptions::get().set<bool>( "ALWAYS_USE_CMD", false );

        } else {
            XmlOptions::get().setRaw( "ADJUST_TIME_CMD", command );
            XmlOptions::write();
        }

        update();

        return;

    }


    //____________________________________________________________________
    void ClockWidget::_toggleInverseColors( bool value )
    {

        Debug::Throw( "ClockWidget::_toggleInverseColors" );
        MainWindow::_toggleInverseColors( value );

        // update internal colors
        if( XmlOptions::get().get<bool>( "USE_DEFAULT_COLORS" ) )
        {
            _setHourColor( foregroundColor() );
            _setMinuteColor( foregroundColor() );
            _setSecondColor( foregroundColor() );
        }

        update();
    }

    //__________________________________________________________________________
    void ClockWidget::_checkAlarms( void )
    {


        // retrieve the data and time from now
        QDateTime current( QDateTime::currentDateTime().addSecs( 3600*delay_ ) );

        QDate current_date( current.date() );
        QTime current_time( current.time() );
        Alarm::List& alarms( alarms_.alarms() );
        for( Alarm::List::iterator iter = alarms.begin(); iter != alarms.end(); ++iter )
        {
            if( !iter->triggered( current_date, current_time ) ) continue;

            // update window with current time
            // this is needed for when the method is called prior to updating the clock display
            // in which case there appear to be a mismatch between the two.
            update();

            // create and raise alarm dialog
            AlarmDialog* dialog( new AlarmDialog( 0, *iter ) );
            connect( dialog, SIGNAL(repeat(Alarm)), SLOT(_setRepeat(Alarm)) );
            connect( dialog, SIGNAL(closed(Alarm)), SLOT(_unsetRepeat(Alarm)) );
            QtUtil::centerOnDesktop( dialog )->show();
        }

        // restart timer in a minute.
        alarmTimer_.start( 1000*(60-current_time.second() ), this );
        update();

        return;

    }

    //_______________________________________________________
    QPolygonF ClockWidget::_hand( const int& width, const int& length )
    {
        QPolygonF out;
        out
            << QPointF(-width, -length)
            << QPointF( width, -length)
            << QPointF( width, -width)
            << QPointF(-width, -width)
            << QPointF(-width, -length)
            ;

        return out;
    }

}
