#ifndef _DiskLoad_h_
#define _DiskLoad_h_

// $Id: DiskLoad.h,v 1.15 2013/04/29 18:27:15 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/


#include "Counter.h"
#include "Data.h"
#include "DataSamples.h"
#include "Debug.h"
#include "DiskId.h"
#include "QOrderedSet.h"

#include <QDateTime>
#include <QHash>

namespace DISK
{

    //! measure relative cpu load from /proc/stat file
    class DiskLoad: public Counter
    {

        public:

        //! constructor
        DiskLoad( void );

        //! convert number into string with proper unit
        static QString toString( double value );

        //! update data from filesystems
        virtual void update( void ) = 0;

        //! load disk_id to string map
        virtual void loadIdMap( void ) = 0;

        //! devices
        typedef QOrderedSet<QString> DeviceSet;

        //! get device names
        virtual DeviceSet devices( void );

        //! stores mounted disks
        virtual void updateMountedDevices( void );

        //! returns true if device is mounted
        virtual bool isMounted( const QString& device ) const;

        //! shortcut for DataSamples set
        typedef QHash< DiskId, DataSamples > SampleMap;

        //! short cut for diskid to name map
        typedef QHash< QString, DiskId > IdMap;

        //! changes sampleSize_
        void setSampleSize( int value )
        {
            sampleSize_ = value;
            for( SampleMap::iterator iter = sampleMap_.begin(); iter != sampleMap_.end(); iter++ )
            { iter.value().setSampleSize( value ); }
        }

        //! changes sample size
        void setDataSize( int value )
        {
            dataSize_ = value;
            for( SampleMap::iterator iter = sampleMap_.begin(); iter != sampleMap_.end(); iter++ )
            { iter.value().setDataSize( value ); }
        }

        //! retrieve data for a given cpu name
        DataSamples::DataList data( const QString& name ) const
        {
            IdMap::const_iterator iter( idMap_.find( name ) );
            if( iter != idMap_.end() ) return data( iter.value() );
            return DataSamples::DataList();
        }

        //! retrieve data for a given cpu name
        DataSamples::DataList data( const DiskId& id ) const
        {
            SampleMap::const_iterator iter( sampleMap_.find( id ) );
            if( iter == sampleMap_.end() ) return DataSamples::DataList();
            else return iter.value().data();
        }

        //! used to sort data according to selected data type
        class LessFTor
        {

            public:

            //! constructor
            LessFTor( const Data::Types& type ):
                type_( type )
                {}

            //! predicate
            bool operator()( const Data& first, const Data& second ) const
            { return double(first.sum( type_ ))/first.time() < double( second.sum( type_ ))/second.time(); }

            private:

            //! data type
            Data::Types type_;

        };

        protected:

        //! sample size
        virtual int _sampleSize( void ) const
        { return sampleSize_; }

        //! data size
        virtual int _dataSize( void ) const
        { return dataSize_; }

        //! sample map
        virtual const SampleMap& _sampleMap( void ) const
        { return sampleMap_; }

        //! sample map
        virtual SampleMap& _sampleMap( void )
        { return sampleMap_; }

        //! file name
        virtual const QString& _mountFileName( void ) const
        { return mountFilename_; }

        //! devices
        virtual const DeviceSet& _mountedDevices( void ) const
        { return mountedDevices_; }

        //! disk ids
        virtual const IdMap& _diskIds( void ) const
        { return idMap_; }

        //! disk ids
        virtual IdMap& _diskIds( void )
        { return idMap_; }

        //! timer
        virtual const QTime& _timer( void ) const
        { return timer_; }

        private:

        //! number of samples to average on
        int sampleSize_;

        //! number of data points to store
        int dataSize_;

        //! data samples
        SampleMap sampleMap_;

        //! file from where mounted devices are read
        const QString mountFilename_;

        //! mounted device
        DeviceSet mountedDevices_;

        //! disk id to string map
        IdMap idMap_;

        //! time
        QTime timer_;

    };

};

#endif
