// $Id: ConfigurationDialog.cpp,v 1.47 2014/01/10 12:14:16 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "ConfigurationDialog.h"
#include "ConfigurationDialog.moc"

#include "GridLayout.h"
#include "IconEngine.h"
#include "Icons.h"
#include "OptionAnchorWidget.h"
#include "OptionCheckBox.h"
#include "OptionSpinBox.h"
#include "SvgConfiguration.h"
#include "TransparencyConfiguration.h"

#include <QGroupBox>
#include <QLabel>
#include <QLayout>
#include <QPushButton>

//_________________________________________________________
ConfigurationDialog::ConfigurationDialog( QWidget* parent ):
BaseConfigurationDialog( parent )
{

    Debug::Throw( "ConfigurationDialog::ConfigurationDialog.\n" );

    // base configuration
    baseConfiguration( 0, Base|List|Animations);

    // transparency configuration
    QWidget* page;
    TRANSPARENCY::TransparencyConfiguration* transparency_configuration;

    page = &addPage( IconEngine::get( ICONS::PREFERENCE_TRANSPARENCY ), tr( "Transparency" ), tr( "Window transparency settings" ) );
    page->layout()->addWidget( transparency_configuration = new TRANSPARENCY::TransparencyConfiguration( page ));
    addOptionWidget( transparency_configuration );

    // svg background
    page = &addPage( IconEngine::get( ICONS::PREFERENCE_BACKGROUND ), tr( "Background" ), tr( "Window background settings" ) );
    SVG::SvgConfiguration* svg_configuration( new SVG::SvgConfiguration( page ) );
    page->layout()->addWidget( svg_configuration);
    addOptionWidget( svg_configuration );

    // generic objects
    QGroupBox *box;
    OptionCheckBox *checkbox;
    OptionSpinBox* spinbox;
    QLabel* label;

    // window
    page = &addPage( IconEngine::get( ICONS::PREFERENCE_MAIN_WINDOW ), tr( "Main Window" ), tr( "Window geometry and behavior" ) );
    page->layout()->addWidget( box = new QGroupBox( tr( "Anchor" ), page ));
    box->setLayout( new QVBoxLayout() );
    box->layout()->setSpacing(5);

    OptionAnchorWidget* anchor = new OptionAnchorWidget( box, "ANCHOR" );
    box->layout()->addWidget( anchor );
    addOptionWidget( anchor );

    page->layout()->addWidget( box = new QGroupBox( tr( "Sizes" ), page ));

    QGridLayout* gridLayout = new QGridLayout();
    gridLayout->setSpacing(5);
    box->setLayout( gridLayout );

    gridLayout->addWidget( label = new QLabel( tr( "Geometry:" ), box ), 0, 0, 1, 1 );
    label->setAlignment( Qt::AlignVCenter|Qt::AlignRight );
    gridLayout->addWidget( spinbox = new OptionSpinBox( box, "WINDOW_WIDTH", false ), 0, 1, 1, 1 );
    spinbox->setSuffix( tr( "px" ) );
    spinbox->setMinimum( 5 );
    spinbox->setMaximum( 2048 );
    spinbox->setToolTip( tr( "Main window width (pixels)" ) );
    addOptionWidget( spinbox );
    label->setBuddy( spinbox );


    gridLayout->addWidget( label = new QLabel( "x", box ), 0, 2, 1, 1 );
    gridLayout->addWidget( spinbox = new OptionSpinBox( box, "WINDOW_HEIGHT" ), 0, 3, 1, 1 );
    spinbox->setSuffix( tr( "px" ) );
    spinbox->setMinimum( 5 );
    spinbox->setMaximum( 2048 );
    spinbox->setToolTip( tr( "Main window height (pixels)" ) );
    addOptionWidget( spinbox );
    label->setBuddy( spinbox );

    gridLayout->addWidget( label = new QLabel( tr( "Margin:" ), box ), 1, 0, 1, 1 );
    label->setAlignment( Qt::AlignVCenter|Qt::AlignRight );

    gridLayout->addWidget( spinbox = new OptionSpinBox( box, "WINDOW_BORDER" ), 1, 1, 1, 3 );
    spinbox->setSuffix( tr( "px" ) );
    spinbox->setMinimum( 0 );
    spinbox->setMaximum( 100 );
    spinbox->setToolTip( tr( "Distance (pixels) between window edge and text/histogram" ) );
    addOptionWidget( spinbox );
    label->setBuddy( spinbox );

    gridLayout->addWidget( label = new QLabel( tr( "Grid:" ), box ), 2, 0, 1, 1 );
    label->setAlignment( Qt::AlignVCenter|Qt::AlignRight );

    gridLayout->addWidget( spinbox = new OptionSpinBox( box, "GRID" ), 2, 1, 1, 3  );
    spinbox->setSuffix( tr( "px" ) );
    spinbox->setMinimum( 1 );
    spinbox->setMaximum( 50 );
    spinbox->setToolTip(
        tr( "Window position grid size (pixels)\n"
        "(allows easier alignment between \n"
        "the different transparent windows)" ) );
    addOptionWidget( spinbox );
    label->setBuddy( spinbox );

    // options
    page->layout()->addWidget( box = new QGroupBox( tr( "Options" ), page ));
    QVBoxLayout *boxLayout = new QVBoxLayout();
    boxLayout->setSpacing(5);
    box->setLayout( boxLayout );

    boxLayout->addWidget( checkbox = new OptionCheckBox( tr( "Keep above others" ), box, "STAYS_ON_TOP" ));
    checkbox->setToolTip( tr( "Keep window above all others" ) );
    addOptionWidget( checkbox );

    boxLayout->addWidget( checkbox = new OptionCheckBox( tr( "Sticky" ), box, "STICKY" ));
    checkbox->setToolTip( tr( "Make window appear on all desktops" ) );
    addOptionWidget( checkbox );

    boxLayout->addWidget( checkbox = new OptionCheckBox( tr( "Hide from taskbar" ), box, "HIDE_FROM_TASKBAR" ));
    checkbox->setToolTip( tr( "Hide window from taskbar" ) );
    addOptionWidget( checkbox );

    boxLayout->addWidget( checkbox = new OptionCheckBox( tr( "Round corners" ), this, "ROUND_CORNERS" ) );
    checkbox->setToolTip( tr( "Round/square window corners" ) );
    addOptionWidget( checkbox );

    boxLayout->addWidget( checkbox = new OptionCheckBox( tr( "Lock position" ), box, "LOCK_POSITION" ));
    checkbox->setToolTip( tr( "Lock the window position on desktop" ) );
    addOptionWidget( checkbox );

    read();

}
