// $Id: GeometryApplication.cpp,v 1.25 2014/01/15 14:01:52 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "GeometryApplication.h"
#include "GeometryApplication.moc"

#include "ApplicationType.h"
#include "BaseConfigurationDialog.h"
#include "GeometryDialog.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QApplication>

//____________________________________________
GeometryApplication::GeometryApplication( CommandLineArguments arguments ):
    BaseApplication( 0, arguments ),
    Counter( "GeometryApplication" ),
    mainWindow_( 0 )
{}

//____________________________________________
GeometryApplication::~GeometryApplication( void )
{

    Debug::Throw( "GeometryApplication::~GeometryApplication.\n" );
    if( mainWindow_ ) delete mainWindow_;

}

//____________________________________________
bool GeometryApplication::realizeWidget( void )
{
    Debug::Throw( "GeometryApplication::realizeWidget.\n" );

    // check if the method has already been called.
    if( !BaseApplication::realizeWidget() ) return false;

    // top widget
    mainWindow_ = new GeometryDialog();
    connect( mainWindow_, SIGNAL(geometryRequested()), SLOT(_sendGeometryRequest()) );
    connect( mainWindow_, SIGNAL(geometryChanged(ApplicationData)), SLOT(_sendGeometry(ApplicationData)) );

    // show main window
    emit configurationChanged();
    mainWindow().show();
    mainWindow().reload();

    return true;

}

//____________________________________________
void GeometryApplication::usage( void ) const
{
    _usage( "transparency-settings" );
    commandLineParser().usage();
    return;
}

//___________________________________________________________
void GeometryApplication::_configuration( void )
{
    Debug::Throw( "GeometryApplication::_configuration.\n" );

    // save configuration
    emit saveConfiguration();

    BaseConfigurationDialog dialog(0, BaseConfigurationDialog::Default );
    dialog.setWindowTitle( "Configuration - Transparency-settings" );

    connect( &dialog, SIGNAL(configurationChanged()), SIGNAL(configurationChanged()) );
    dialog.centerOnWidget( qApp->activeWindow() );
    dialog.exec();
}

//_________________________________________________
void GeometryApplication::_sendGeometryRequest( void )
{

    Debug::Throw( "GeometryApplication::_sendGeometryRequest.\n" );
    _sendRequest( SERVER::ServerCommand::GeometryRequest );

}

//_________________________________________________
void GeometryApplication::_sendGeometry( ApplicationData data )
{

    Debug::Throw() << "GeometryApplication::_sendGeometry - " << data << endl;

    // send geometry
    SERVER::ApplicationId id( _applicationManager().id() );
    id.setName( data.name() );

    // create command
    QRect rect( data.geometry() );
    SERVER::ServerCommand command( id, SERVER::ServerCommand::Geometry );
    CommandLineArguments arguments;
    arguments
        << "transparency "
        << QString().setNum( rect.left() )
        << QString().setNum( rect.top())
        << QString().setNum( rect.width() )
        << QString().setNum( rect.height() )
        << QString().setNum( data.order() );
    command.setArguments( arguments );

    // send to server
    _applicationManager().client().sendCommand( command );

}

//_________________________________________________
void GeometryApplication::_sendRequest( SERVER::ServerCommand::CommandType command_type )
{

    Debug::Throw( "GeometryApplication::_sendRequest.\n" );

    // loop over application types
    for( ApplicationType::Map::const_iterator iter = ApplicationType::names().begin(); iter != ApplicationType::names().end(); ++iter )
    {

        Debug::Throw() << "Application::_sendRequest - sending to: " << iter.value() << endl;

        // create command that match name
        SERVER::ApplicationId id( _applicationManager().id() );
        id.setName( iter.value() );

        // create command
        SERVER::ServerCommand command( id, command_type );
        _applicationManager().client().sendCommand( command );

    }

}

//________________________________________________
bool GeometryApplication::_processCommand( SERVER::ServerCommand command )
{


    // try base class
    if( BaseApplication::_processCommand( command ) ) return true;

    // test command type
    switch( command.command() )
    {

        case SERVER::ServerCommand::Raise:
        if( mainWindow_ ) mainWindow().uniconify();
        return true;

        case SERVER::ServerCommand::Geometry:
        if( mainWindow_ )
        {

            // parse command
            CommandLineParser parser( commandLineParser( command.arguments() ) );
            QStringList arguments( parser.orphans() );
            if( arguments.size() >= 5 )
            {
                ApplicationData data;
                data.setName( arguments[0] );
                data.setGeometry( QRect(
                    QPoint( arguments[1].toInt(), arguments[2].toInt() ),
                    QSize( arguments[3].toInt(), arguments[4].toInt() ) ) );

                if( arguments.size() > 5 ) data.setOrder( arguments[5].toInt() );
                if( arguments.size() > 6 ) data.setIcon( QIcon( arguments[6] ));

                Debug::Throw() << "GeometryApplication::_processCommand - " << data << endl;

                mainWindow().add( data );
            }

        }
        return true;

        default: return false;
    }
}
