// $Id: PictureApplication.cpp,v 1.43 2014/01/15 14:01:52 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "PictureApplication.h"
#include "PictureApplication.moc"

#include "ImageFileDialog.h"
#include "InformationDialog.h"
#include "PictureListDialog.h"
#include "PictureWidget.h"
#include "Icons.h"
#include "IconEngine.h"
#include "PictureParser.h"
#include "QtUtil.h"
#include "Util.h"
#include "XmlOptions.h"
#include "Config.h"

#include <QApplication>

//____________________________________________
PictureApplication::PictureApplication( CommandLineArguments arguments ) :
    BaseApplication( 0, arguments ),
    Counter( "PictureApplication" )
{
    Debug::Throw( "PictureApplication::PictureApplication.\n" );
    qApp->setQuitOnLastWindowClosed( false );
}

//____________________________________________
PictureApplication::~PictureApplication( void )
{ Debug::Throw( "PictureApplication::~PictureApplication.\n" ); }

//____________________________________________
bool PictureApplication::realizeWidget( void )
{
    Debug::Throw( "PictureApplication::realizeWidget.\n" );

    // check if the method has already been called.
    if( !BaseApplication::realizeWidget() ) return false;

    // actions
    _installActions();

    // read images from arguments
    QStringList filenames( CommandLineParser().parse( _arguments() ).orphans() );

    // if no images are provided, use the ones stored in resource file
    if( filenames.isEmpty() ) _readPictures();
    else {

        foreach( const QString& filename, filenames )
        {

            File file = File( filename ).expand();

            // check file existence
            if( !file.exists() ) continue;
            Debug::Throw(0) << "PictureApplication::realizeWidget - loading: " << file << endl;

            PICTURES::PictureAttributes attributes;
            attributes.setFilename( file );
            PICTURES::PictureWidget* widget = new PICTURES::PictureWidget( 0, attributes );
            widget->adjustSizeToImage();
            QtUtil::centerOnDesktop( widget );
            _addPicture( widget );
        }

    }

    // configuration
    emit configurationChanged();

    // if no picture is set, creates an empty one
    if( pictureWidgets_.empty() )
    {

        PICTURES::PictureWidget* widget( new PICTURES::PictureWidget() );
        _addPicture( widget );
        widget->show();

    } else {

        bool hasVisible( false );
        for( PICTURES::PictureWidget::List::const_iterator iter = pictureWidgets_.begin(); iter != pictureWidgets_.end(); ++iter )
        {
            if( (*iter)->attributes().isVisible() )
            {
                hasVisible = true;
                (*iter)->show();
            }
        }

        // if no picture is visible, raise the window list dialog instead
        if( !hasVisible )
        {
            InformationDialog( 0, tr( "All registered pictures are presently hidden.\n Select the one you want to show from the list." ) ).centerOnWidget( qApp->activeWindow() ).exec();
            _showPictureList();
        }

    }

    return true;

}

//____________________________________________
void PictureApplication::usage( void ) const
{
    _usage( "transparent-pictures" );
    commandLineParser().usage();
    return;
}

//_______________________________________________
QString PictureApplication::applicationName( void ) const
    { return "Picture Viewer"; }

//_______________________________________________
QString PictureApplication::applicationVersion( void ) const
{ return VERSION; }

//_______________________________________________
QString PictureApplication::buildTimeStamp( void ) const
{ return BUILD_TIMESTAMP; }

//_______________________________________________
void PictureApplication::_configuration( void )
{}

//_______________________________________________
void PictureApplication::_raiseAll( void )
{
    Debug::Throw( "PictureApplication::_raiseAll.\n" );

    for( PICTURES::PictureWidget::List::const_iterator iter = pictureWidgets_.begin(); iter != pictureWidgets_.end(); ++iter )
    { if( ( *iter )->attributes().isVisible() ) (*iter)->uniconify(); }

}

//_______________________________________________
void PictureApplication::_updateConfiguration( void )
{
    showFileNameAction_->setChecked( XmlOptions::get().get<bool>( "SHOW_FILENAME" ) );
    keepAspectRatioAction_->setChecked( XmlOptions::get().get<bool>( "KEEP_ASPECT_RATIO" ) );
}

//________________________________________________
bool PictureApplication::_processCommand( SERVER::ServerCommand command )
{

    Debug::Throw( "PictureApplication::_processCommand.\n" );
    if( BaseApplication::_processCommand( command ) ) return true;
    if( command.command() == SERVER::ServerCommand::Raise )
    {
        for( PICTURES::PictureWidget::List::const_iterator iter = pictureWidgets_.begin(); iter != pictureWidgets_.end(); ++iter )
        { (*iter)->uniconify(); }

        return true;
    } else return false;

}

//_______________________________________________
void PictureApplication::_writePictures( bool check )
{

    Debug::Throw( "PictureApplication::_writePictures.\n" );

    // if taskBars have not been initialized, do nothing
    if( !_realized() ) return;

    // check taskBar
    if( pictureWidgets_.empty() ) return;

    // try open XML file
    File file( XmlOptions::get().raw( "RC_FILE" ) );

    PICTURES::PictureParser parser;
    for( PICTURES::PictureWidget::List::const_iterator iter = pictureWidgets_.begin(); iter != pictureWidgets_.end(); ++iter )
    {
        if( (*iter)->attributes().filename() != PICTURES::PictureAttributes::NO_PICTURE )
        { parser.addAttributes( (*iter)->attributes() ); }
    }

    bool written( parser.write( file ) );
    if( (!written) && check )
    {
        const QString buffer = QString( tr( "Unable to write to file '%1'" ) ).arg( file );
        InformationDialog( 0, buffer ).exec();
    }

    return;

}

//_______________________________________________________________________________________________
void PictureApplication::_newPicture( void )
{
    Debug::Throw( "PictureApplication::_newPicture.\n" );

    // create dialog
    ImageFileDialog dialog(0);
    dialog.setFileMode( QFileDialog::ExistingFile );
    dialog.setAcceptMode( QFileDialog::AcceptOpen );
    QtUtil::centerOnWidget( &dialog, qApp->activeWindow() );

    // try get current picture widget
    PICTURES::PictureWidget* current( qobject_cast<PICTURES::PictureWidget*>( qApp->activeWindow() ) );
    if( current && current->attributes().filename() != PICTURES::PictureAttributes::NO_PICTURE )
    {
        File path( File( current->attributes().filename() ).path() );
        if( path.exists() && path.isDirectory() ) dialog.setDirectory( path );
    }

    if( dialog.exec() == QDialog::Rejected ) return;

    // retrieve selected files
    QStringList files( dialog.selectedFiles() );

    // create attributes
    PICTURES::PictureAttributes attributes;

    // copy size from current widget
    if( current && current->attributes().size().isValid() )
    {
        attributes.setSize( current->attributes().size() );
        attributes.setPosition( QtUtil::centerOnDesktop( attributes.size() ) );
    }

    // check file size
    if( files.size() == 1 )
    { attributes.setFilename( files.front() ); }

    // create new picture and save in list
    PICTURES::PictureWidget* widget( new PICTURES::PictureWidget( 0, attributes ) );
    _addPicture( widget );
    widget->show();

    return;

}

//_______________________________________________________________________________________________
void PictureApplication::_showPictureList( void )
{
    Debug::Throw( "PictureApplication::_showPictureList.\n" );

    PICTURES::PictureAttributes::List attributes;
    for( PICTURES::PictureWidget::List::const_iterator iter = pictureWidgets_.begin(); iter != pictureWidgets_.end(); ++iter )
    { attributes << (*iter)->attributes(); }

    PICTURES::PictureListDialog dialog( 0, attributes );
    if( dialog.centerOnWidget( qApp->activeWindow() ).exec() )
    {

        // retrieve attributes from dialog
        attributes = dialog.attributes();

        // delete widgets that are not found in the new list of attributes
        PICTURES::PictureWidget::List tmpWidgets;
        for( PICTURES::PictureWidget::List::iterator iter( pictureWidgets_.begin() ); iter != pictureWidgets_.end(); ++iter )
        {
            PICTURES::PictureAttributes::List::iterator attributes_iter( std::find_if( attributes.begin(), attributes.end(), PICTURES::PictureWidget::SamePositionFTor( (*iter)->attributes().position() ) ) );
            if( attributes_iter == attributes.end() )
            {

                (*iter)->deleteLater();

            } else {

                // update with attributes
                (*iter)->setAttributes( *attributes_iter );
                tmpWidgets << *iter;

                if( (*iter)->attributes().isVisible() ) (*iter)->show();
                else (*iter)->hide();

                // remove attributes from list
                attributes.erase( std::remove( attributes.begin(), attributes.end(), *attributes_iter ), attributes.end() );

            }

        }

        // update widget list
        pictureWidgets_ = tmpWidgets;

        // add remaining attributes
        for( PICTURES::PictureAttributes::List::iterator iter = attributes.begin(); iter != attributes.end(); ++iter )
        {
            PICTURES::PictureWidget* widget( new PICTURES::PictureWidget( 0, *iter ) );
            _addPicture( widget );
            if( widget->attributes().isVisible() ) widget->show();
        }

    }

    // check if at least one widget is visible
    for( PICTURES::PictureWidget::List::iterator iter( pictureWidgets_.begin() ); iter != pictureWidgets_.end(); ++iter )
    { if( (*iter)->attributes().isVisible() ) return; }

    // quit if all windows are hidden
    qApp->quit();

    return;

}

//_______________________________________________
void PictureApplication::_pictureDestroyed( QObject* object )
{

    Debug::Throw( "PictureApplication::_pictureDestroyed.\n" );

    // remove object from list of widgets
    pictureWidgets_.erase( std::remove( pictureWidgets_.begin(), pictureWidgets_.end(), object ), pictureWidgets_.end() );

    // quit if no pictures remain
    if( pictureWidgets_.empty() ) qApp->quit();

}

//________________________________________________
void PictureApplication::_toggleKeepAspectRatio( bool state )
{

    Debug::Throw( "PictureApplication::_toggleKeepAspectRatio.\n" );
    XmlOptions::get().set<bool>( "KEEP_ASPECT_RATIO", state );

}

//________________________________________________
void PictureApplication::_toggleShowFilename( bool state )
{

    Debug::Throw( "PictureApplication::_toggleShowFilename.\n" );
    XmlOptions::get().set<bool>( "SHOW_FILENAME", state );

}

//_______________________________________________
void PictureApplication::_pictureHidden( void )
{
    Debug::Throw( "PictureApplication::_pictureHidden.\n" );

    for( PICTURES::PictureWidget::List::const_iterator iter = pictureWidgets_.begin(); iter != pictureWidgets_.end(); ++iter )
    { if( ( *iter )->attributes().isVisible() ) return; }

    qApp->quit();

}

//_______________________________________________
void PictureApplication::_addPicture( PICTURES::PictureWidget* widget )
{
    Debug::Throw( "PictureApplication::_addPicture.\n" );

    connect( widget, SIGNAL(pictureHidden()), SLOT(_pictureHidden()) );
    connect( widget, SIGNAL(destroyed(QObject*)), SLOT(_pictureDestroyed(QObject*)) );

    widget->addAction( &closeAction() );

    pictureWidgets_ << widget;

}

//_______________________________________________
void PictureApplication::_readPictures( void )
{
    Debug::Throw( "PictureApplication::_readPictures.\n" );

    // read tasks from db file
    const File file( XmlOptions::get().raw("RC_FILE" ) );
    PICTURES::PictureParser parser;
    parser.read( file );

    // check errors
    if( parser.xmlError() )
    {
        QString buffer = QString( tr( "Errors occured while parsing file '%1':\n" ) ).arg( file );
        buffer += parser.xmlError().toString();
        InformationDialog( 0, buffer ).exec();
    }

    // clear picture windows
    for( PICTURES::PictureWidget::List::iterator iter = pictureWidgets_.begin(); iter != pictureWidgets_.end(); ++iter )
    { (*iter)->deleteLater(); }
    pictureWidgets_.clear();

    const PICTURES::PictureAttributes::List& attributes( parser.attributes() );
    for( PICTURES::PictureAttributes::List::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter )
    { _addPicture( new PICTURES::PictureWidget( 0, *iter ) ); }

    return;

}

//_______________________________________________________________________________________________
void PictureApplication::_installActions( void )
{

    Debug::Throw( "PictureApplication::_installActions.\n" );

    // rename about action
    aboutAction().setText( "About Picture Viewer" );

    // configuration
    connect( this, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( qApp, SIGNAL(aboutToQuit()), SLOT(_writePictures()) );

    // raise all
    raiseAllAction_ = new QAction( IconEngine::get( ICONS::RAISE_ALL ), tr( "Raise All" ), this );
    raiseAllAction_->setToolTip( tr( "Raise all transparent windows" ) );
    connect( raiseAllAction_, SIGNAL(triggered()), SLOT(_raiseAll()) );

    // new picture action
    newPictureAction_ = new QAction( IconEngine::get( ICONS::ADD ), tr( "New" ), this );
    newPictureAction_->setToolTip( tr( "Add new picture" ) );
    newPictureAction_->setShortcut( QKeySequence::New );
    connect( newPictureAction_, SIGNAL(triggered()), SLOT(_newPicture()) );

    // picture list
    pictureListAction_ = new QAction( IconEngine::get( ICONS::FIND ), tr( "Picture List" ), this );
    pictureListAction_->setToolTip( tr( "Show list of pictures" ) );
    connect( pictureListAction_, SIGNAL(triggered()), SLOT(_showPictureList()) );

    keepAspectRatioAction_ = new QAction( tr( "Keep Aspect Ratio" ), this );
    keepAspectRatioAction_->setCheckable( true );
    keepAspectRatioAction_->setChecked( false );
    connect( keepAspectRatioAction_, SIGNAL(toggled(bool)), SLOT(_toggleKeepAspectRatio(bool)) );

    showFileNameAction_ = new QAction( tr( "Show Filename" ), this );
    showFileNameAction_->setCheckable( true );
    showFileNameAction_->setChecked( false );
    connect( showFileNameAction_, SIGNAL(toggled(bool)), SLOT(_toggleShowFilename(bool)) );

}
