
// $Id: TransparentWindow.cpp,v 1.93 2013/11/15 14:50:40 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TransparentWindow.h"
#include "TransparentWindow.moc"

#include "Anchor.h"
#include "BaseApplication.h"
#include "BaseContextMenu.h"
#include "Color.h"
#include "CompositeEngine.h"
#include "ConfigurationDialog.h"
#include "CustomPixmap.h"
#include "DebugMenu.h"
#include "DefaultHelpText.h"
#include "HelpManager.h"
#include "Icons.h"
#include "IconEngine.h"
#include "IconSize.h"
#include "OptionAnchorWidget.h"
#include "QtUtil.h"
#include "RoundedRegion.h"
#include "Singleton.h"
#include "SvgEngine.h"
#include "SvgHelpText.h"
#include "TransparencyHelpText.h"
#include "WinUtil.h"
#include "XmlOptions.h"
#include "X11Util.h"

#include <QApplication>
#include <QDesktopWidget>
#include <QPainter>

// all desktop flags (used for stickyness)
const unsigned long ALL_DESKTOPS = ~0;

//_________________________________________________________________
TransparentWindow::TransparentWindow( QWidget* parent ):
    TRANSPARENCY::TransparentWidget( parent, Qt::FramelessWindowHint ),
    menuSeparator_( 0 ),
    direction_( X11Util::_NET_WM_MOVERESIZE_MOVE ),
    button_( Qt::NoButton ),
    anchor_( Anchor::Default ),
    order_( -1 ),
    grid_( 1 )
{
    Debug::Throw( "TransparentWindow::TransparentWindow.\n" );

    // actions
    _installActions();
    _updateConfiguration();

    // help system
    help_ = new BASE::HelpManager( this );
    help_->install( TRANSPARENCY::helpText );
    help_->install( SVG::helpText );
    help_->install( BASE::helpText, false );

    setProperty( "_kde_no_blur_behind", true );

    connect( &SVG::SvgEngine::get(), SIGNAL(changed()), SLOT(_updateConfiguration()) );
    connect( &SVG::SvgEngine::get(), SIGNAL(changed()), SLOT(setBackgroundChanged()) );
    connect( qApp->desktop(), SIGNAL(resized(int)), SLOT(_desktopResized()) );

}

//_____________________________________________________
void TransparentWindow::uniconify( void )
{ QtUtil::uniconify( this ); }

//________________________________________________
void TransparentWindow::processOption( const XmlOption& option )
{

    Debug::Throw()
        << "TransparentWindow::processOption -"
        << " recieved option: " << option.name()
        << " value: " << option.raw()
        << endl;

    QAction* action(0);
    if( option.name() == "STAYS_ON_TOP" ) action = staysOnTopAction_;
    else if( option.name() == "LOCK_POSITION" ) action = lockPositionAction_;
    else if( option.name() == "HIDE_FROM_TASKBAR" ) action = hideFromTaskBarAction_;
    else if( option.name() == "STICKY" ) action = stickyAction_;
    else if( option.name() == "ROUND_CORNERS" ) action = roundCornersAction_;
    else if( option.name() == "USE_SVG" ) action = useSvgAction_;
    else if( option.name() == "TRANSPARENCY_INVERSE_COLORS" ) action = &inverseColorsAction();
    else Debug::Throw(0) << "TransparentWindow::processOption - option " << option.name() << " not implemented" << endl;

    if( action ) action->setChecked( option.get<bool>() );

}

//________________________________________________
QRect TransparentWindow::innerGeometry( void ) const
{ return _outerPadding().adjustedRect( geometry() ); }

//________________________________________________
void TransparentWindow::setInnerGeometry( const QRect& rect )
{
    const BASE::Margins outerPadding( _outerPadding() );
    setGeometry( rect.adjusted( -outerPadding.left(), -outerPadding.top(), outerPadding.right(), outerPadding.bottom() ) );
}

//________________________________________________
void TransparentWindow::_installContextMenuActions( BaseContextMenu* menu )
{

    Debug::Throw( "TransparentWindow::_installContextMenuActions.\n" );

    menuSeparator_ = menu->addSeparator();
    menu->addAction( &_configurationAction() );

    {
        QMenu* flagsMenu = menu->QMenu::addMenu( tr( "Flags" ) );
        flagsMenu->addAction( staysOnTopAction_ );
        flagsMenu->addAction( hideFromTaskBarAction_ );
        flagsMenu->addAction( stickyAction_ );
        flagsMenu->addAction( lockPositionAction_ );
        flagsMenu->addAction( roundCornersAction_ );
        flagsMenu->addAction( useSvgAction_ );
        flagsMenu->addAction( &inverseColorsAction() );
        flagsMenu->addSeparator();
        flagsMenu->addAction( lockFlagsAction_ );
    }

    {
        QMenu *helpMenu = menu->QMenu::addMenu( tr( "Help" ) );
        helpMenu->addAction( &help_->displayAction() );
        #if defined(Q_WS_X11) || defined(Q5_WS_X11)
        helpMenu->addAction( &reloadBlurRegionAction() );
        #endif
        helpMenu->addAction( reloadBackgroundAction_ );
        helpMenu->addSeparator();

        helpMenu->addAction( &Singleton::get().application<BaseApplication>()->aboutAction() );
        helpMenu->addAction( &Singleton::get().application<BaseApplication>()->aboutQtAction() );

    }

    // debug menu
    DebugMenu *debugMenu( new DebugMenu( menu ) );
    menu->addAction( debugMenu->menuAction() );
    debugMenu->menuAction()->setVisible( false );

    menu->addSeparator();
    menu->addAction( &Singleton::get().application<BaseApplication>()->closeAction() );

    return;

}

//________________________________________________
void TransparentWindow::_configureFromDialog( ConfigurationDialog& dialog )
{

    Debug::Throw( "TransparentWindow::_configureFromDialog.\n" );
    connect( &dialog, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( &dialog, SIGNAL(configurationChanged()), Singleton::get().application(), SIGNAL(configurationChanged()) );
    dialog.centerOnParent();
    dialog.exec();

}

//___________________________________________________
void TransparentWindow::timerEvent( QTimerEvent* event )
{
    if( event->timerId() == desktopResizeTimer_.timerId() )
    {

        // get position
        QPoint position( this->pos() );

        Debug::Throw()
            << "TransparentWindow::timerEvent -"
            << " " << windowTitle()
            << " position: (" << position.x() << "," << position.y() << ")"
            << " target: (" << position_.x() << "," << position_.y() << ")"
            << endl;

        if( position != position_ )
        {

            move( position_ );

        } else desktopResizeTimer_.stop();

    } else {

        // base class implementation
        return TRANSPARENCY::TransparentWidget::timerEvent( event );

    }

}

//___________________________________________________
void TransparentWindow::moveEvent( QMoveEvent* event )
{
    geometryTimer_.start( 200, this );
    TransparentWidget::moveEvent( event );
    if( grid_ > 1 )
    {

        QPoint point( _gridPoint( event->pos() ) );
        if( point != position_ )
        {
            position_ = point;
            move( point );
        }

    }
}

//___________________________________________________
void TransparentWindow::resizeEvent( QResizeEvent *event )
{
    Debug::Throw( "TransparentWindow::resizeEvent.\n" );
    geometryTimer_.start( 200, this );
    _updateMask();

    if( TRANSPARENCY::CompositeEngine::get().isAvailable() )
    { _updateBlurRegion( _outerPadding().adjustedRect( rect() ).translated( mapTo( window(), QPoint(0,0) ) ) ); }

    TransparentWidget::resizeEvent( event );
    WinUtil( this ).enableBlurBehind( _outerPadding() );
}

//___________________________________________________
void TransparentWindow::mousePressEvent( QMouseEvent *event )
{

    TRANSPARENCY::TransparentWidget::mousePressEvent( event );

    _setButton( event->button() );
    if( _button() == Qt::LeftButton && !lockPositionAction_->isChecked() )
    {
        // check modifiers
        if( event->modifiers() & Qt::ShiftModifier ) direction_ = _direction( event->pos() );
        else direction_ = X11Util::_NET_WM_MOVERESIZE_MOVE;

        clickPosition_ = event->pos() + QPoint(geometry().topLeft() - frameGeometry().topLeft());
        clickSize_ = size();

        // try use X11 Window Manager to handle event
        if( X11Util::get().moveResizeWidget( this, event->globalPos(), direction_ ) ) return;
        else {

            // otherwise use curstom
            if( event->modifiers() & Qt::ShiftModifier )
            {
                setCursor( ( direction_ == X11Util::_NET_WM_MOVERESIZE_SIZE_TOPLEFT || direction_ == X11Util::_NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT ) ? Qt::SizeFDiagCursor : Qt::SizeBDiagCursor );
            } else setCursor( Qt::SizeAllCursor );

        }

    }

}

//___________________________________________________
void TransparentWindow::mouseReleaseEvent( QMouseEvent* )
{
    _setButton( Qt::NoButton );
    direction_ = X11Util::_NET_WM_MOVERESIZE_MOVE;
    unsetCursor();
}

//___________________________________________________
void TransparentWindow::mouseMoveEvent( QMouseEvent *event )
{

    if( _button() == Qt::LeftButton && !lockPositionAction_->isChecked() )
    {

        QPoint global_point( _gridPoint( event->globalPos() - clickPosition_ ) );
        if( direction_ == X11Util::_NET_WM_MOVERESIZE_MOVE ) move( global_point );
        else {

            QRect newRect( geometry() );
            QPoint point( _gridPoint( event->pos() - clickPosition_ ) );
            if( direction_ == X11Util::_NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT )
            {

                newRect.setWidth( clickSize_.width() + point.x() );
                newRect.setHeight( clickSize_.height() + point.y() );

            } else if( direction_ == X11Util::_NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT ) {

                newRect.setLeft( global_point.x() );
                newRect.setHeight( clickSize_.height() + point.y() );

            } else if( direction_ == X11Util::_NET_WM_MOVERESIZE_SIZE_TOPRIGHT ) {

                newRect.setWidth( clickSize_.width() + point.x() );
                newRect.setTop( global_point.y() );

            } else {

                newRect.setLeft( global_point.x() );
                newRect.setTop( global_point.y() );

            }

            setGeometry( newRect );

        }

    }
}

//___________________________________________________
void TransparentWindow::mouseDoubleClickEvent( QMouseEvent *event )
{
    TRANSPARENCY::TransparentWidget::mouseDoubleClickEvent( event );
    _setButton( event->button() );
    return;
}

//________________________________________________
void TransparentWindow::contextMenuEvent( QContextMenuEvent* event )
{

    Debug::Throw( "TransparentWindow::contextMenuEvent.\n" );

    BaseContextMenu menu( this );
    menu.setHideDisabledActions( true );
    _installContextMenuActions( &menu );
    menu.exec( event->globalPos() );

}


//________________________________________________
void TransparentWindow::_configuration( void )
{

    // save configuration
    Debug::Throw( "TransparentWindow::_configuration.\n" );
    ConfigurationDialog dialog;
    _configureFromDialog( dialog );

}

//________________________________________________
void TransparentWindow::_updateConfiguration( void )
{

    Debug::Throw( "TransparentWindow::_updateConfiguration.\n" );

    // position
    if( XmlOptions::get().contains( "ANCHOR" ) ) anchor_ = XmlOptions::get().get<unsigned int>( "ANCHOR" );

    // grid
    grid_ = XmlOptions::get().get<int>( "GRID" );

    // actions
    staysOnTopAction_->setChecked( XmlOptions::get().get<bool>( "STAYS_ON_TOP" ) );
    lockPositionAction_->setChecked( XmlOptions::get().get<bool>( "LOCK_POSITION" ) );
    roundCornersAction_->setChecked( XmlOptions::get().get<bool>( "ROUND_CORNERS" ) );

    // otherwise toggle the action directly
    stickyAction_->setChecked( XmlOptions::get().get<bool>( "STICKY" ) );
    hideFromTaskBarAction_->setChecked( XmlOptions::get().get<bool>( "HIDE_FROM_TASKBAR" ) );

    // reload svg
    useSvgAction_->setChecked( XmlOptions::get().get<bool>( "USE_SVG" ) );
    if( useSvgAction_->isChecked() && SVG::SvgEngine::get().reload() )
    { setBackgroundChanged( true ); }

    // margins
    _clearMargins();
    BASE::Margins margins;
    if( useSvgAction_->isChecked() && SVG::SvgEngine::get().isValid() )
    {
        margins = SVG::SvgEngine::get().margins();
        margins -= BASE::Margins( 4 );
    }

    // add option
    margins += BASE::Margins( XmlOptions::get().get<int>( "WINDOW_BORDER" ) );
    _setMargins( margins );

    // outer padding
    _clearOuterPadding();
    if( useSvgAction_->isChecked() && SVG::SvgEngine::get().isValid() )
    {
        // read either outer padding or margins from SVG, and adjust
        BASE::Margins outerPadding( SVG::SvgEngine::get().outerPadding() );
        if( outerPadding.isNull() )
        {
            outerPadding = SVG::SvgEngine::get().margins();
            if( !outerPadding.isNull() ) outerPadding -= BASE::Margins( 4 );
        }

        _setOuterPadding( outerPadding );

    }

    // reload background
    #if !defined(Q_WS_X11) && !defined(Q5_WS_X11)
    reloadBackgroundAction_->setEnabled( useSvgAction_->isChecked() );
    #endif

    // flags locking
    lockFlagsAction_->setChecked( XmlOptions::get().get<bool>( "LOCK_FLAGS" ) );

    // order
    setOrder( XmlOptions::get().get<bool>( "ORDER" ) );

}

//________________________________________________
void TransparentWindow::_toggleStaysOnTop( bool state )
{
    Debug::Throw( "TransparentWindow::_toggleStaysOnTop.\n" );

    // check that widget is top level
    if( parentWidget() ) return;

    #if defined(Q_WS_X11) || defined(Q5_WS_X11)

    // on X11 one uses NET_WM directly
    X11Util::get().changeProperty( this, X11Util::_NET_WM_STATE_STAYS_ON_TOP, state );
    X11Util::get().changeProperty( this, X11Util::_NET_WM_STATE_ABOVE, state );

    #else

    bool visible( !isHidden() );
    if( visible )
    {
        hide();
        qApp->processEvents();
    }

    // Qt implementation
    if( state ) setWindowFlags( windowFlags() | Qt::WindowStaysOnTopHint );
    else setWindowFlags( windowFlags() & ~Qt::WindowStaysOnTopHint );

    if( visible ) show();

    #endif

    XmlOptions::get().set<bool>( "STAYS_ON_TOP", state );
    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "STAYS_ON_TOP", XmlOptions::get().option( "STAYS_ON_TOP" ) ) ); }

}

//________________________________________________
void TransparentWindow::_toggleLockPosition( bool state )
{

    XmlOptions::get().set<bool>( "LOCK_POSITION",  state );
    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "LOCK_POSITION", XmlOptions::get().option( "LOCK_POSITION" ) ) ); }

}

//________________________________________________
void TransparentWindow::_toggleHideFromTaskBar( bool state )
{

    Debug::Throw() << "TransparentWindow::_toggleHideFromTaskBar - state: " << state << endl;

    // check that widget is top level
    if( parentWidget() ) return;

    #if defined(Q_WS_X11) || defined(Q5_WS_X11)
    X11Util::get().changeProperty( this, X11Util::_NET_WM_STATE_SKIP_TASKBAR, state );
    X11Util::get().changeProperty( this, X11Util::_NET_WM_STATE_SKIP_PAGER, state );
    #endif

    #if defined(Q_OS_WIN)
    WinUtil( this ).toggleHideFromTaskBar( state );
    #endif

    // update option
    XmlOptions::get().set<bool>( "HIDE_FROM_TASKBAR", state );
    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "HIDE_FROM_TASKBAR", XmlOptions::get().option( "HIDE_FROM_TASKBAR" ) ) ); }

    return;

}

//________________________________________________
void TransparentWindow::_toggleSticky( bool state )
{

    Debug::Throw() << "TaskBarBase::_toggleSticky - state: " << state << endl;

    // check that widget is top level
    if( parentWidget() ) return;

    #if defined(Q_WS_X11) || defined(Q5_WS_X11)
    if( X11Util::get().isSupported( X11Util::_NET_WM_STATE_STICKY ) )
    {

        // change property depending on state
        X11Util::get().changeProperty( this, X11Util::_NET_WM_STATE_STICKY, state );

    } else if( X11Util::get().isSupported( X11Util::_NET_WM_DESKTOP ) ) {

        unsigned long desktop = X11Util::get().cardinal( X11Util::get().appRootWindow(), X11Util::_NET_CURRENT_DESKTOP );
        X11Util::get().changeCardinal( this, X11Util::_NET_WM_DESKTOP, state ? X11Util::ALL_DESKTOPS:desktop );

    }

    #endif

    // update option
    XmlOptions::get().set<bool>( "STICKY", state );
    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "STICKY", XmlOptions::get().option( "STICKY" ) ) ); }

    return;

}

//________________________________________________
void TransparentWindow::_toggleRoundCorners( bool state )
{

    _updateMask();

    // only update options if action is enabled
    XmlOptions::get().set<bool>( "ROUND_CORNERS", state );
    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "ROUND_CORNERS", XmlOptions::get().option( "ROUND_CORNERS" ) ) ); }

    return;

}

//________________________________________________
void TransparentWindow::_toggleSvg( bool state )
{

    if( state && !SVG::SvgEngine::get().isValid() ) SVG::SvgEngine::get().reload();

    // only update options if action is enabled
    if( useSvgAction_->isEnabled() )
    {
        XmlOptions::get().set<bool>( "USE_SVG", state );
        if( lockFlagsAction_->isChecked() )
        { emit optionChanged( XmlOption( "USE_SVG", XmlOptions::get().option( "USE_SVG" ) ) ); }
    }

    // update background
    setBackgroundChanged( true );
    update();

    return;
}


//____________________________________________________________________
void TransparentWindow::_toggleInverseColors( bool value )
{

    Debug::Throw( "TransparentWindow::_toggleInverseColors" );
    TRANSPARENCY::TransparentWidget::_toggleInverseColors( value );

    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "TRANSPARENCY_INVERSE_COLORS", XmlOptions::get().option( "TRANSPARENCY_INVERSE_COLORS" ) ) ); }

    update();
}


//____________________________________________________________
void TransparentWindow::_toggleLockFlags( bool state )
{

    XmlOptions::get().set<bool>( "LOCK_FLAGS", state );
    return;

}

//________________________________________________
void TransparentWindow::_updateBackgroundPixmap( void )
{

    TransparentWidget::_updateBackgroundPixmap();
    Debug::Throw()
        << "TransparentWindow::_updateBackgroundPixmap -"
        << " size: " << _backgroundPixmap().width() << "," << _backgroundPixmap().height()
        << endl;

    if( useSvgAction_->isChecked() && SVG::SvgEngine::get().isValid() )
    {
        QPainter painter( &_backgroundPixmap() );
        painter.setRenderHints(QPainter::SmoothPixmapTransform);
        painter.setCompositionMode(QPainter::CompositionMode_Source );
        painter.drawPixmap( QPoint( 0, 0 ), SVG::SvgEngine::get().get( size() ) );
        painter.end();
    }

}

//________________________________________________
void TransparentWindow::_desktopResized( void )
{

    Debug::Throw(0) << "TransparentWindow::_desktopResized - " << windowTitle() << endl;

    // position
    if( XmlOptions::get().contains( "WINDOW_X" ) && XmlOptions::get().contains( "WINDOW_Y" ) )
    { _setPosition( QPoint( XmlOptions::get().get<int>( "WINDOW_X" ), XmlOptions::get().get<int>( "WINDOW_Y" ) ) ); }

    XmlOptions::write();
}

//________________________________________________
void TransparentWindow::_installActions( void )
{
    Debug::Throw( "TransparentWindow::_installActions.\n" );

    // configure
    configurationAction_ = new QAction( IconEngine::get( ICONS::CONFIGURE ), tr( "Configuration" ), this );
    configurationAction_->setToolTip( tr( "Application configuration" ) );
    connect( configurationAction_, SIGNAL(triggered()), SLOT(_configuration()) );

    // stays on top
    addAction( staysOnTopAction_ = new QAction( tr( "Keep Above" ), this ) );
    staysOnTopAction_->setToolTip( tr( "Keep window above all others" ) );
    staysOnTopAction_->setCheckable( true );
    staysOnTopAction_->setChecked( windowFlags() & Qt::WindowStaysOnTopHint );
    connect( staysOnTopAction_, SIGNAL(toggled(bool)), SLOT(_toggleStaysOnTop(bool)) );

    // lock
    addAction( lockPositionAction_ = new QAction( tr( "Lock Position" ), this ) );
    lockPositionAction_->setCheckable( true );
    lockPositionAction_->setEnabled( windowFlags() & Qt::FramelessWindowHint );
    connect( lockPositionAction_, SIGNAL(toggled(bool)), SLOT(_toggleLockPosition(bool)) );

    // hide from taskBar
    addAction( hideFromTaskBarAction_ = new QAction( tr( "Hide from TaskBar" ), this ) );
    hideFromTaskBarAction_->setCheckable( true );
    hideFromTaskBarAction_->setChecked( false );
    connect( hideFromTaskBarAction_, SIGNAL(toggled(bool)), SLOT(_toggleHideFromTaskBar(bool)) );

    // sticky
    addAction( stickyAction_ = new QAction( tr( "Sticky" ), this ) );
    stickyAction_->setToolTip( tr( "Make window appear on all desktops" ) );
    stickyAction_->setCheckable( true );
    stickyAction_->setChecked( false );
    connect( stickyAction_, SIGNAL(toggled(bool)), SLOT(_toggleSticky(bool)) );

    // round corners
    addAction( roundCornersAction_ = new QAction( tr( "Round Corners" ), this ) );
    roundCornersAction_->setToolTip( tr( "Round window corners" ) );
    roundCornersAction_->setCheckable( true );
    roundCornersAction_->setChecked( false );
    roundCornersAction_->setEnabled( windowFlags() & Qt::FramelessWindowHint );
    connect( roundCornersAction_, SIGNAL(toggled(bool)), SLOT(_toggleRoundCorners(bool)) );

    // svg
    addAction( useSvgAction_ = new QAction( tr( "Use Svg" ), this ) );
    useSvgAction_->setToolTip( tr( "Use svg for background" ) );
    useSvgAction_->setCheckable( true );
    useSvgAction_->setChecked( false );
    connect( useSvgAction_, SIGNAL(toggled(bool)), SLOT(_toggleSvg(bool)) );

    // reload background
    addAction( reloadBackgroundAction_ = new QAction( IconEngine::get( ICONS::RELOAD ), tr( "Reload Background" ), this ) );
    reloadBackgroundAction_->setToolTip( tr( "Reload window background" ) );
    reloadBackgroundAction_->setShortcut( QKeySequence::Refresh );
    connect( reloadBackgroundAction_, SIGNAL(triggered()), &SVG::SvgEngine::get(), SLOT(forceReload()) );

    #if defined(Q_WS_X11) || defined(Q5_WS_X11)
    // also connect to blur background
    connect( reloadBackgroundAction_, SIGNAL(triggered()), &reloadBlurRegionAction(), SLOT(trigger()) );
    #endif

    // lock
    addAction( lockFlagsAction_ = new QAction( tr( "Lock Flags" ), this ) );
    lockFlagsAction_->setToolTip( tr( "Use common flags for all transparency applications" ) );
    lockFlagsAction_->setCheckable( true );
    lockFlagsAction_->setChecked( false );
    connect( lockFlagsAction_, SIGNAL(toggled(bool)), SLOT(_toggleLockFlags(bool)) );

    #if defined(Q_WS_X11) || defined(Q5_WS_X11)
    hideFromTaskBarAction_->setEnabled( X11Util::get().isSupported( X11Util::_NET_WM_STATE_SKIP_TASKBAR ) );
    stickyAction_->setEnabled(
        X11Util::get().isSupported( X11Util::_NET_WM_STATE_STICKY ) ||
        X11Util::get().isSupported( X11Util::_NET_WM_DESKTOP ) );
    #endif

}

//________________________________________________
void TransparentWindow::_setPosition( QPoint point )
{

    // get desktop screen geometry
    QRect screen = qApp->desktop()->screenGeometry( qApp->desktop()->screenNumber( this ) );

    Debug::Throw()
        << "TransparentWindow::_setPosition -"
        << " position: " << point.x() << "," << point.y()
        << " padding: " << _outerPadding().topLeft().x() << "," << _outerPadding().topLeft().y()
        << " screen: " << screen.width() << "," << screen.height()
        << endl;

    if( anchor_ & Anchor::Right ) point.setX( screen.width() - point.x() );
    if( anchor_ & Anchor::HorizontalCenter ) point.setX( screen.width()/2 + point.x() );

    if( anchor_ & Anchor::Bottom ) point.setY( screen.height() - point.y() );
    if( anchor_ & Anchor::VerticalCenter ) point.setY( screen.height()/2 + point.y() );

    // add outer padding
    point -= _outerPadding().topLeft();

    // update position if needed
    position_ = point;
    if( point != this->pos() )
    {

        Debug::Throw()
            << "TransparentWindow::_setPosition -"
            << " " << windowTitle()
            << " target: (" << position_.x() << "," << position_.y() << ")"
            << endl;

        move( point );

        // start desktop resize timer
        desktopResizeTimer_.start( 500, this );

    }

}

//________________________________________________
QPoint TransparentWindow::_position( void ) const
{

    Debug::Throw( "TransparentWindow::_position.\n" );
    QPoint point( mapToGlobal( QPoint(0,0) ) );
    point += _outerPadding().topLeft();

    QRect screen = qApp->desktop()->screenGeometry( qApp->desktop()->screenNumber( this ) );

    if( anchor_ & Anchor::Right ) point.setX( screen.width() - point.x() );
    if( anchor_ & Anchor::HorizontalCenter ) point.setX( point.x() - screen.width()/2 );

    if( anchor_ & Anchor::Bottom ) point.setY( screen.height() - point.y() );
    if( anchor_ & Anchor::VerticalCenter ) point.setY( point.y() - screen.height()/2 );

    Debug::Throw()
        << "TransparentWindow::_position -"
        << " position: " << point.x() << "," << point.y()
        << " padding: " << _outerPadding().topLeft().x() << "," << _outerPadding().topLeft().y()
        << " screen: " << screen.width() << "," << screen.height()
        << endl;

    return point;

}

//________________________________________________
void TransparentWindow::_updateMask( void )
{
    if( roundCornersAction_->isChecked() ) setMask( RoundedRegion( rect() ) );
    else clearMask();
}

//________________________________________________
QPoint TransparentWindow::_gridPoint( QPoint point ) const
{

    return point;

    if( grid_ <= 1 ) return point;

    QPoint out( grid_*int( 0.5+ point.x()/grid_ ), grid_*int( 0.5 + point.y()/grid_ ) );
    while( out.x() - point.x() > grid_/2 ) out.rx() -= grid_;
    while( out.x() - point.x() <= -grid_/2 ) out.rx() += grid_;

    while( out.y() - point.y() > grid_/2 ) out.ry() -= grid_;
    while( out.y() - point.y() <= -grid_/2 ) out.ry() += grid_;

    return out;

}

//________________________________________________
X11Util::Direction TransparentWindow::_direction( const QPoint& point ) const
{
    const int halfWidth( width()/2 );
    const int halfHeight( height()/2 );
    if( point.x() < halfWidth ) return point.y() < halfHeight ?  X11Util::_NET_WM_MOVERESIZE_SIZE_TOPLEFT: X11Util::_NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT;
    else return point.y() < halfHeight ? X11Util::_NET_WM_MOVERESIZE_SIZE_TOPRIGHT: X11Util::_NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT;
}

//__________________________________________________________
void TransparentWindow::Color::update( void )
{

    Debug::Throw( "TransparentWindow::Color::update.\n" );
    QColor color( XmlOptions::get().get<BASE::Color>(name_) );
    if( !color.isValid() ) color = qApp->palette().color( QPalette::WindowText );

    ( foreground_ = color ).setAlpha( XmlOptions::get().get<int>( "TRANSPARENCY_FOREGROUND_INTENSITY" ) );
    ( background_ = color ).setAlpha( XmlOptions::get().get<int>( "TRANSPARENCY_BACKGROUND_INTENSITY" ) );

    // update gradient
    const int lightenAmount( 130 );
    useGradient_ = XmlOptions::get().get<bool>( "TRANSPARENCY_USE_GRADIENT" );
    foregroundGradient_ = QLinearGradient( foregroundGradient_.start(), foregroundGradient_.finalStop() );
    foregroundGradient_.setColorAt(0, foreground_.lighter( lightenAmount ) );
    foregroundGradient_.setColorAt(1, foreground_ );

    backgroundGradient_ = QLinearGradient( backgroundGradient_.start(), backgroundGradient_.finalStop() );
    backgroundGradient_.setColorAt(0, background_.lighter( lightenAmount ) );
    backgroundGradient_.setColorAt(1, background_ );

}

//__________________________________________________________
void TransparentWindow::Color::updateGradient( const QSize& size )
{
    Debug::Throw( "TransparentWindow::Color::updateGradient.\n" );
    foregroundGradient_.setFinalStop( 0, size.height() );
    backgroundGradient_.setFinalStop( 0, size.height() );
}

//__________________________________________________________
QIcon TransparentWindow::Color::icon( void ) const
{
    Debug::Throw( "TransparentWindow::Color::icon.\n" );

    QPixmap pixmap( CustomPixmap().empty( IconSize( IconSize::Large ) ) );
    QPainter painter( &pixmap );
    painter.setRenderHints(QPainter::Antialiasing );
    painter.setPen( foreground_ );
    painter.setBrush( background_ );

    QRectF rect = QRect( QPoint(0,0), IconSize( IconSize::Large ) ).adjusted( 2, 2, -2, -2 );
    painter.drawEllipse( rect );
    painter.end();

    return QIcon( pixmap );
}
