// $Id: MemoryLoad.cpp,v 1.18 2013/04/29 18:27:17 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

/*!
  \file    MemoryLoad.cpp
  \brief   measures relative memory load from /proc/stat file
  \author  Hugo Pereira
  \version $Revision: 1.18 $
  \date    $Date: 2013/04/29 18:27:17 $
*/

#include "MemoryLoad.h"
#include "Debug.h"

#include <QFile>
#include <QStringList>
#include <QTextStream>

namespace MEMORY
{

    //________________________________________________
    MemoryLoad::MemoryLoad( void ):
        Counter( "MemoryLoad" ),
        version_( MAIN::LINUX_26 ),
        filename_( "/proc/meminfo" ),
        sampleSize_( 1 ),
        dataSize_( 1 )
    {}

    //________________________________________________
    void MemoryLoad::update( void )
    {
        Data current( _data() );
        sampleList_ << current;

        if( sampleList_.size() > sampleSize_ )
        {

            sampleList_.removeFirst();

            Data sum;
            for( DataList::iterator iter = sampleList_.begin(); iter != sampleList_.end(); iter ++ )
                if( ! sum.valid() ) sum = (*iter);
            else sum += (*iter);
            sum.divide( sampleSize_ );
            dataList_ << sum;

            if( dataList_.size() > dataSize_ ) dataList_.removeFirst();

        }

        return;

    }

    //________________________________________________
    MemoryLoad::Data MemoryLoad::_data( void )
    {



        Data out;

        // open file
        QFile in( filename_ );
        if( !in.open( QIODevice::ReadOnly ) )
        {
            Debug::Throw( "MemoryLoad::_data - cannot open file.\n" );
            return MemoryLoad::Data();
        }

        switch (version_)
        {
            case MAIN::LINUX_24:
            {
                bool found_mem( false );
                bool found_swap( false );
                QStringList contents( QString(in.readAll()).split( "\n" ) );
                for( QStringList::const_iterator iter = contents.begin(); iter != contents.end(); ++iter )
                {
                    QString line( *iter );
                    if( line.indexOf( "Mem:" ) >= 0 )
                    {

                        QString dummy;
                        QTextStream stream( &line );
                        stream >> dummy >> out.totalMemory() >> out.memory();
                        found_mem = (stream.status() == QTextStream::Ok );
                    }

                    if( line.indexOf( "Swap:" ) >= 0 )
                    {
                        QString dummy;
                        QTextStream stream( &line );
                        stream >> dummy >> out.totalSwap() >> out.swap();
                        found_swap = (stream.status() == QTextStream::Ok );
                    }

                    if( found_mem && found_swap ) break;

                }
                out.valid() = ( found_mem && found_swap );

            }
            break;

            case MAIN::LINUX_26:
            {
                bool found_tot_mem( false );
                bool found_free_mem( false );
                bool found_tot_swap( false );
                bool found_free_swap( false );
                unsigned long tot_mem( 0 );
                unsigned long free_mem( 0 );
                unsigned long tot_swap( 0 );
                unsigned long free_swap( 0 );

                QStringList contents( QString(in.readAll()).split( "\n" ) );
                for( QStringList::const_iterator iter = contents.begin(); iter != contents.end(); ++iter )
                {
                    QString line( *iter );
                    if( line.indexOf( "MemTotal:" ) >= 0 )
                    {
                        QString dummy;
                        QTextStream stream( &line );
                        stream >> dummy >> tot_mem;
                        found_tot_mem = (stream.status() == QTextStream::Ok );
                    }

                    if( line.indexOf( "SwapTotal:" ) >= 0 )
                    {
                        QString dummy;
                        QTextStream stream( &line );
                        stream >> dummy >> tot_swap;
                        found_tot_swap = (stream.status() == QTextStream::Ok );
                    }

                    if( line.indexOf( "MemFree:" ) >= 0 )
                    {
                        QString dummy;
                        QTextStream stream( &line );
                        stream >> dummy >> free_mem;
                        found_free_mem = (stream.status() == QTextStream::Ok );
                    }

                    if( line.indexOf( "SwapFree:" ) >= 0 )
                    {
                        QString dummy;
                        QTextStream stream( &line );
                        stream >> dummy >> free_swap;
                        found_free_swap = (stream.status() == QTextStream::Ok );
                    }

                    if( found_tot_mem && found_tot_swap && found_free_mem && found_free_swap ) break;

                }

                out.valid() = ( found_tot_mem && found_tot_swap && found_free_mem && found_free_swap );
                if( out.valid() )
                {

                    out.memory() = tot_mem - free_mem;
                    out.totalMemory() = tot_mem;
                    out.swap() = tot_swap - free_swap;
                    out.totalSwap() = tot_swap;

                }
            }
            break;

            default:
            Debug::Throw(0) << "MemoryLoad::_data - unsuported kernel version " << endl;
            break;

        }

        // divide out to avoid overflows
        out = out.divide(1000);

        return out;
    }

}
