// $Id: NetworkWidget.cpp,v 1.61 2014/01/12 18:36:25 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "NetworkWidget.h"
#include "NetworkWidget.moc"

#include "BaseContextMenu.h"
#include "IconSize.h"
#include "InformationDialog.h"
#include "NetworkConfigurationDialog.h"
#include "ToolTipWidget.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QTextStream>
#include <algorithm>

namespace NETWORK
{

    //__________________________________________________________
    NetworkWidget::NetworkWidget( QWidget *parent ):
        HistogramWidget( parent ),
        networkMonitor_( this ),
        displayMask_( NetworkLoad::ALL )
    {

        Debug::Throw( "NETWORK::NetworkWidget::NetworkWidget.\n" );
        setWindowTitle( Util::windowTitle( tr( "Network Monitor" ) ) );

        colorMap_.insert( NetworkLoad::RECIEVED, Color( "RECIEVED_COLOR" ) );
        colorMap_.insert( NetworkLoad::TRANSMITTED, Color( "TRANSMITTED_COLOR" ) );

        _installActions();
        _configurationAction().setText( tr( "Network Monitor Settings..." ) );
        _updateConfiguration();

    }

    //______________________________________________________________________
    void NetworkWidget::_installContextMenuActions( BaseContextMenu* menu )
    {
        Debug::Throw( "NETWORK::NetworkWidget::_installContextMenuActions.\n" );

        // create parent menu
        HistogramWidget::_installContextMenuActions( menu );

        // network menu
        menu->insertSeparator( &_menuSeparator() );
        foreach( QAction* action, displayActionGroup_->actions() )
        { menu->insertAction( &_menuSeparator(), action ); }

        menu->insertMenu( &_menuSeparator(), &networkMonitor_.menu() );

    }

    //______________________________________________________________________
    bool NetworkWidget::_updateToolTip( void )
    {

        Debug::Throw( "NETWORK::NetworkWidget::_updateToolTip.\n" );

        _toolTipWidget().setPixmap( windowIcon().pixmap( IconSize( IconSize::Maximum ) ) );

        // html formated tooltip
        // title
        QString buffer;
        QTextStream what( &buffer );
        what << "<div align='center' style='border: 10px;'><b>" << windowTitle() << "</b></div>" << endl;

        // header
        what << "<table style='margin: 10px;' >" << endl;
        what << "<tr><td align='right'>host: </td><td>" << _hostName() << "</td></tr>" << endl;
        what << "<tr><td align='right'>port: </td><td>" << networkMonitor_.device() << "</td></tr>" << endl;
        what << endl;

        // data
        const NetworkLoad::DataList& data( networkMonitor_.data() );
        if( !data.empty() ) {
            what << "<tr><td align='right'>transmission: </td><td>" << NetworkLoad::toString( double(data.back().bytesTransmitted())/data.back().time() ) << "</td></tr>" << endl;
            what << "<tr><td align='right'>reception: </td><td>" << NetworkLoad::toString( double(data.back().bytesRecieved())/data.back().time() ) << "</td></tr>" << endl;
        } else what << "<tr><td colspan='2'>acquiring</td></tr>" << endl;
        what << "</table>"<< endl;

        _toolTipWidget().setText( buffer );
        return _toolTipWidget().isHidden();

    }

    //______________________________________________________________________
    void NetworkWidget::timerEvent( QTimerEvent* event )
    {

        if( event->timerId() == _updateTimer().timerId() )
        {

            networkMonitor_.update();
            update();

        } else return HistogramWidget::timerEvent( event );

    }

    //______________________________________________________________________
    void NetworkWidget::_paint( QPaintDevice& device, const QRect& clip )
    {

        // check border against window size
        if( !_margins().adjustedRect( rect() ).isValid() )
        {
            InformationDialog( this, tr( "Window size is too small with respect to margins.\n Margins set to 0." ) ).exec();
            _setMargins(0);
        }

        // store size
        QRect rect( _margins().adjustedRect( this->rect() ) );
        const int width( rect.width() );
        const int height( rect.height() );

        // create painter
        QPainter painter( &device );
        painter.setClipRect( clip );
        painter.setRenderHints(QPainter::Antialiasing);

        // retrieve data
        NetworkLoad::DataList data( networkMonitor_.data() );
        while( data.size() > width ) data.removeFirst();

        // store flags locally to spare time
        if( !data.empty() && displayMask_ )
        {

            const bool showRecieved( displayMask_ & NetworkLoad::RECIEVED );
            const bool showTransmitted( displayMask_ & NetworkLoad::TRANSMITTED );

            // running points definition
            double xMin( _margins().left() + 1 );
            double x( xMin );
            double yMin( height + _margins().bottom() + 1 );
            double yMax( yMin );

            QPolygonF recievedPolygon;
            QPolygonF transmittedPolygon;

            // autoscale painting
            double scale( _getScale( data ) );
            for( NetworkLoad::DataList::const_iterator iter = data.begin(); iter != data.end(); ++iter )
            {

                yMax = yMin;

                if( showRecieved )
                {
                    yMax -= qMax( 1.0, scale*iter->bytesRecieved()/iter->time() );
                    recievedPolygon << QPointF( x, yMax );
                }

                if( showTransmitted )
                {
                    yMax -= qMax( 1.0, scale*iter->bytesTransmitted()/iter->time() );
                    transmittedPolygon << QPointF( x, yMax );
                }

                x+=1;
            }

            // close polygons
            /*
            polygons are manually 'subtracted' because it should work quicker than
            the general case Qt implementation
            */
            if( showTransmitted )
            {
                if( showRecieved ) {
                    for( int i = recievedPolygon.size()-1; i >= 0; i-- )
                    { transmittedPolygon << recievedPolygon[i]; }
                } else {
                    transmittedPolygon << QPointF( x-1, yMin );
                    transmittedPolygon << QPointF( xMin, yMin );
                }

                transmittedPolygon << transmittedPolygon.front();
            }

            if( showRecieved )
            {
                recievedPolygon << QPointF( x-1, yMin );
                recievedPolygon << QPointF( xMin, yMin );
                recievedPolygon << recievedPolygon.front();
            }

            // prepare pen
            QPen pen;
            pen.setJoinStyle( Qt::RoundJoin );
            pen.setWidthF( 1.5 );

            // draw transmitted polygon
            if( showTransmitted )
            {
                pen.setColor( colorMap_[NetworkLoad::TRANSMITTED].foreground() );
                painter.setPen( pen );
                painter.setBrush( colorMap_[NetworkLoad::TRANSMITTED].background() );
                painter.drawPolygon( transmittedPolygon );

            }

            // draw recieved polygon
            if( showRecieved )
            {
                pen.setColor( colorMap_[NetworkLoad::RECIEVED].foreground() );
                painter.setPen( pen );
                painter.setBrush( colorMap_[NetworkLoad::RECIEVED].background() );
                painter.drawPolygon( recievedPolygon );
            }

        }

        // draw load
        if( _showLoadAction().isChecked() || data.empty() )
        { _paintLoad( painter, rect ); }

        // draw host
        _paintHost( painter, rect );

        painter.end();

    }

    //______________________________________________________________________
    void NetworkWidget::resizeEvent( QResizeEvent* event )
    {
        int width = qMax( 0, event->size().width() - _margins().width() );
        networkMonitor_.setDataSize( width );
        colorMap_[NetworkLoad::RECIEVED].updateGradient( event->size() );
        colorMap_[NetworkLoad::TRANSMITTED].updateGradient( event->size() );
        return HistogramWidget::resizeEvent( event );
    }

    //________________________________________________
    void NetworkWidget::_configuration( void )
    {

        Debug::Throw( "NETWORK::NetworkWidget::_configuration.\n" );

        NetworkConfigurationDialog dialog;
        _configureFromDialog( dialog );

    }

    //__________________________________________________________
    void NetworkWidget::_updateConfiguration( void )
    {
        Debug::Throw( "NETWORK::NetworkWidget::_updateConfiguration.\n" );

        // colors
        colorMap_[NetworkLoad::RECIEVED].update();
        colorMap_[NetworkLoad::TRANSMITTED].update();

        // update mask
        displayMask_ = 0;
        if( XmlOptions::get().get<bool>( "SHOW_TRANSMITTED" ) ) displayMask_ |= NetworkLoad::TRANSMITTED;
        if( XmlOptions::get().get<bool>( "SHOW_RECIEVED" ) ) displayMask_ |= NetworkLoad::RECIEVED;
        _updateDisplayActions();

        // base class update
        HistogramWidget::_updateConfiguration();
        update();
        Debug::Throw( "NETWORK::NetworkWidget::_updateConfiguration - done.\n" );

    }

    //________________________________________________
    void NetworkWidget::_toggleDisplay( QAction* action )
    {

        Debug::Throw( "NETWORK::NetworkWidget::_toggleDisplay.\n" );
        NetworkLoad::DataType bit( displayActions_[action] );
        if( action->isChecked() ) displayMask_ |= bit;
        else displayMask_ &= ~bit;

        XmlOptions::get().set<bool>( "SHOW_TRANSMITTED", displayMask_ & NetworkLoad::TRANSMITTED );
        XmlOptions::get().set<bool>( "SHOW_RECIEVED", displayMask_ & NetworkLoad::RECIEVED );

        update();

    }

    //______________________________________________________________________
    void NetworkWidget::_installActions( void )
    {

        Debug::Throw( "NETWORK::NetworkWidget::_installActions.\n" );

        // display
        displayActionGroup_ = new QActionGroup( this );
        displayActionGroup_->setExclusive( false );
        displayActions_.insert( displayActionGroup_->addAction( tr( "Reception" ) ), NetworkLoad::RECIEVED );
        displayActions_.insert( displayActionGroup_->addAction( tr( "Transmission" ) ), NetworkLoad::TRANSMITTED );

        for( DisplayActionMap::iterator iter = displayActions_.begin(); iter != displayActions_.end(); ++iter )
        { iter.key()->setCheckable( true ); }

        _updateDisplayActions();
        connect( displayActionGroup_, SIGNAL(triggered(QAction*)), SLOT(_toggleDisplay(QAction*)) );

    }

    //______________________________________________________________________
    void NetworkWidget::_updateDisplayActions( void )
    {
        Debug::Throw( "NETWORK::NetworkWidget::_updateDisplayActions.\n" );
        for( DisplayActionMap::iterator iter = displayActions_.begin(); iter != displayActions_.end(); ++iter )
        {
            iter.key()->setChecked( displayMask_ & iter.value() );
            iter.key()->setIcon( colorMap_[iter.value()].icon() );
        }

    }

    //_________________________________________________________
    double NetworkWidget::_getScale( const NetworkLoad::DataList& data ) const
    {
        Debug::Throw( "NETWORK::NetworkWidget::_getScale.\n" );

        Q_ASSERT( !data.empty() );

        // get maximum
        NetworkLoad::DataList::const_iterator iter( std::max_element( data.begin(), data.end(), NetworkLoad::LessFTor( displayMask_ ) ) );
        double maximum = double( iter->sum( displayMask_ ) )/iter->time();

        unsigned int height( qMax<unsigned int>( 1, NetworkWidget::height() - _margins().height() ) );
        return (maximum) ? double( height )/maximum:1;
    }

    //_________________________________________________________
    void NetworkWidget::_paintLoad( QPainter& painter, QRect& rect )
    {
        Debug::Throw( "NETWORK::NetworkWidget::_paintLoad.\n" );

        QString buffer;
        NETWORK::NetworkLoad::DataList data( networkMonitor_.data() );
        if( !data.empty() )
        {

            buffer = QString( tr( "Out: %1" ) ).arg( NetworkLoad::toString( double(data.back().bytesTransmitted())/data.back().time() ) );
            _paintText( painter, buffer, rect );

            buffer = QString( tr( "In: %1" ) ).arg( NetworkLoad::toString( double(data.back().bytesRecieved())/data.back().time() ) );
            _paintText( painter, buffer, rect );

        } else {

            buffer = QString( tr( "Acquiring (%1)" ) ).arg( networkMonitor_.device() );
            _paintText( painter, buffer, rect );

        }

        return;

    }

}
