// $Id: PictureWidget.cpp,v 1.45 2013/11/15 14:50:40 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "PictureConfigurationDialog.h"
#include "ImageFileDialog.h"
#include "InformationDialog.h"

#include <QUrl>

#include "PictureWidget.h"
#include "PictureWidget.moc"
#include "QuestionDialog.h"

#include "PictureApplication.h"
#include "BaseContextMenu.h"
#include "Debug.h"
#include "Icons.h"
#include "IconEngine.h"
#include "QtUtil.h"
#include "Singleton.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QDateTime>
#include <QMimeData>
#include <QPainter>
#include <QProcess>

namespace PICTURES
{

    //_______________________________________________
    PictureWidget::PictureWidget( QWidget* parent, const PictureAttributes& attributes ):
        TransparentWindow( parent )
    {

        Debug::Throw( "PictureWidget::PictureWidget.\n" );
        setWindowTitle( Util::windowTitle( tr( "Picture Viewer" ) ) );

        // actions
        _installActions();
        _configurationAction().setText( tr( "Picture Viewer Settings..." ) );

        //! update configuration
        connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
        _updateConfiguration();

        setAcceptDrops( true );
        setAttributes( attributes );

    }

    //________________________________________________
    void PictureWidget::setAttributes( const PictureAttributes& attributes )
    {
        Debug::Throw( "PictureWidget::setAttributes.\n" );
        attributes_ = attributes;

        if( !attributes.size().isEmpty() )
        {

            setGeometry( QRect( attributes.position(), attributes.size() ) );

        } else {

            QSize defaultSize( XmlOptions::get().get<int>( "WINDOW_WIDTH" ), XmlOptions::get().get<int>( "WINDOW_HEIGHT" ) );
            setGeometry( QRect( QtUtil::centerOnDesktop( defaultSize ), defaultSize ) );

        }

        _updatePicture();

    }

    //______________________________________________________________________
    void PictureWidget::adjustSizeToImage( void )
    {

        if( truePixmap_.isNull() ) return;
        resize( truePixmap_.size() + QSize( _margins().width(), _margins().height() ) );

    }

    //______________________________________________________________________
    void PictureWidget::_installContextMenuActions( BaseContextMenu* menu )
    {
        Debug::Throw( "PictureWidget::_installContextMenuActions.\n" );

        // get underlying application
        PictureApplication& application( *Singleton::get().application<PictureApplication>() );

        // get menu from parent
        TransparentWindow::_installContextMenuActions( menu );

        // customize
        menu->insertSeparator( &_menuSeparator() );
        menu->insertAction( &_menuSeparator(), &application.newPictureAction() );
        menu->insertAction( &_menuSeparator(), &_changePictureAction() );
        menu->insertAction( &_menuSeparator(), &_closeAction() );
        menu->insertAction( &_menuSeparator(), &_removePictureAction() );
        menu->insertAction( &_menuSeparator(), &application.pictureListAction() );
        menu->insertSeparator( &_menuSeparator() );
        menu->insertAction( &_menuSeparator(), &application.showFilenameAction() );
        menu->insertAction( &_menuSeparator(), &application.keepAspectRatioAction() );

        return;

    }

    //___________________________________________________
    void PictureWidget::mouseDoubleClickEvent( QMouseEvent *event )
    {
        Debug::Throw( "PictureWidget::mouseDoubleClickEvent.\n" );
        TransparentWindow::mouseDoubleClickEvent( event );
        if( event->button() == Qt::LeftButton )
        { Singleton::get().application<PictureApplication>()->raiseAllAction().trigger(); }

        return;
    }

    //______________________________________________________________________
    void PictureWidget::enterEvent( QEvent* )
    { activateWindow(); }

    //______________________________________________________________________
    void PictureWidget::dragEnterEvent( QDragEnterEvent *event )
    {
        Debug::Throw( "PictureWidget::dragEnterEvent.\n" );
        if (event->mimeData()->hasUrls()) event->acceptProposedAction();
    }

    //______________________________________________________________________
    void PictureWidget::dropEvent( QDropEvent *event )
    {

        Debug::Throw( "PictureWidget::dropEvent.\n" );

        // check if event is valid
        if( !event->mimeData()->hasUrls() ) return;

        // loop over event URLs
        QList<QUrl> urls( event->mimeData()->urls() );
        for( QList<QUrl>::const_iterator iter = urls.begin(); iter != urls.end(); ++iter )
        {
            QFileInfo file_info( iter->toLocalFile() );
            if( file_info.exists() )
            {
                attributes().setFilename( file_info.filePath() );
                _updatePicture();
                event->acceptProposedAction();
            }

        }

        return;

    }

    //______________________________________________________________________
    void PictureWidget::timerEvent( QTimerEvent* event )
    {

        if( event->timerId() == _geometryTimer().timerId() )
        {

            if( !truePixmap_.isNull() )
            {
                _scale();
                update();
            }

            // update attributes
            attributes().setPosition( _position() );
            attributes().setSize( size() );

            TransparentWindow::timerEvent( event );
            _geometryTimer().stop();

        } else TransparentWindow::timerEvent( event );
        return;

    }

    //______________________________________________________________________
    void PictureWidget::_paint( QPaintDevice& device, const QRect& clip )
    {

        if( scaledPixmap_.isNull() )
        {

            if( truePixmap_.isNull() )
            { _paintEmpty( device, clip, shadowOffset() > 0 ); }

        } else {

            // paint scaled image
            QPainter painter( &device );
            painter.setClipRect( clip );
            painter.setClipRect( _margins().adjustedRect( this->rect() ), Qt::IntersectClip );
            painter.setRenderHints(QPainter::Antialiasing );

            double x = 0.5*(width()-scaledPixmap_.width() );
            double y = 0.5*(height()-scaledPixmap_.height());

            if( Singleton::get().application<PictureApplication>()->showFilenameAction().isChecked() ) y -= 0.5*( QFontMetrics( font() ).height() + 0.25*_margins().height() );

            // paint image
            painter.drawPixmap( QPointF( x, y ), scaledPixmap_ );

            painter.end();

            if( Singleton::get().application<PictureApplication>()->showFilenameAction().isChecked() )
            {
                if( shadowOffset() ) _paintFilename( device, clip, true );
                _paintFilename( device, clip, false );
            }

        }

    }

    //________________________________________________
    void PictureWidget::_updateConfiguration( void )
    {

        Debug::Throw( "PictureWidget::_updateConfiguration.\n" );

        // base class update
        TransparentWindow::_updateConfiguration();

        // need to reload background systematically
        setBackgroundChanged( true );
        update();

    }

    //________________________________________________
    void PictureWidget::_scale( void )
    {

        Debug::Throw( "PictureWidget::_scale.\n" );
        if( truePixmap_.isNull() ) return;

        QSize scaled_size( width() - _margins().width(), height() - _margins().height() );

        PictureApplication& application( *Singleton::get().application<PictureApplication>() );
        if( application.showFilenameAction().isChecked() )
        { scaled_size.rheight() -= int( QFontMetrics( font() ).height() + 0.25*_margins().height() ); }

        scaledPixmap_ = truePixmap_.scaled( scaled_size,
            application.keepAspectRatioAction().isChecked() ? Qt::KeepAspectRatio:Qt::IgnoreAspectRatio,
            Qt::SmoothTransformation );

    }

    //________________________________________________
    void PictureWidget::_configuration( void )
    {

        Debug::Throw( "PictureWidget::_configuration.\n" );
        PictureConfigurationDialog dialog;
        connect( &dialog, SIGNAL(configurationChanged()), Singleton::get().application(), SIGNAL(configurationChanged()) );
        dialog.centerOnParent();
        dialog.exec();

    }

    //________________________________________________
    void PictureWidget::_removePicture( void )
    {
        Debug::Throw( "PictureWidget::_removePicture" );
        QuestionDialog dialog( this, tr( "Remove this picture ?" ) );
        dialog.setWindowTitle( tr( "Remove Picture - transparency" ) );
        if( !dialog.centerOnWidget( this ).exec() ) return;
        deleteLater();
        return;

    }

    //________________________________________________
    void PictureWidget::_changePicture( void )
    {

        Debug::Throw( "PictureWidget::_changePicture.\n" );
        ImageFileDialog dialog( this );
        dialog.setFileMode( QFileDialog::ExistingFile );
        dialog.setAcceptMode( QFileDialog::AcceptOpen );
        QtUtil::centerOnParent( &dialog );

        if( attributes().filename() != PictureAttributes::NO_PICTURE )
        {
            File path( File( attributes().filename() ).path() );
            if( path.exists() && path.isDirectory() ) dialog.setDirectory( path );
            dialog.selectFile( attributes().filename() );
        }

        if( dialog.exec() == QDialog::Rejected ) return;

        // retrieve selected files
        QStringList files( dialog.selectedFiles() );

        // check file size
        if( files.size() == 1 )
        {
            attributes().setFilename( files.front() );
            _updatePicture();
        }

        return;

    }

    //________________________________________________
    void PictureWidget::_toggleKeepAspectRatio( bool state )
    {

        Debug::Throw( "PictureWidget::_toggleKeepAspectRatio.\n" );
        _scale();
        update();

    }

    //________________________________________________
    void PictureWidget::_toggleShowFilename( bool state )
    {

        Debug::Throw( "PictureWidget::_toggleShowFilename.\n" );
        _scale();
        update();

    }

    //________________________________________________
    void PictureWidget::_hide( void )
    {
        attributes().setVisible( false );
        hide();
        emit pictureHidden();
    }

    //________________________________________________
    bool PictureWidget::_updatePicture( void )
    {
        Debug::Throw( "PictureWidget::_updatePicture.\n" );

        QString value( attributes().filename() );

        if( value == PictureAttributes::NO_PICTURE )
        {
            truePixmap_ = QPixmap();
            scaledPixmap_ = QPixmap();
            setToolTip( "No Picture" );
            update();
            return false;
        }

        setToolTip( value );
        QPixmap pixmap( value );
        if( pixmap.isNull() )
        {
            QString buffer;
            QTextStream( &buffer ) << "Unable to load image from file " << value;
            InformationDialog( this, buffer ).centerOnWidget( this ).exec();
            return false;
        }

        // store true pixmap
        truePixmap_ = pixmap;

        // scale
        _scale();
        update();
        return true;

    }

    //______________________________________________________________________
    void PictureWidget::_installActions( void )
    {

        Debug::Throw( "PictureWidget::_installActions.\n" );

        PictureApplication& application( *Singleton::get().application<PictureApplication>() );
        addAction( &application.newPictureAction() );

        connect( &application.keepAspectRatioAction(), SIGNAL(toggled(bool)), SLOT(_toggleKeepAspectRatio(bool)) );
        connect( &application.showFilenameAction(), SIGNAL(toggled(bool)), SLOT(_toggleShowFilename(bool)) );

        addAction( removePictureAction_ = new QAction( IconEngine::get( ICONS::REMOVE ), tr( "Remove" ), this ) );
        removePictureAction_->setToolTip( tr( "Remove current picture" ) );
        connect( removePictureAction_, SIGNAL(triggered()), SLOT(_removePicture()) );

        addAction( changePictureAction_ = new QAction( IconEngine::get( ICONS::RELOAD ), tr( "Change" ), this ) );
        changePictureAction_->setToolTip( tr( "Change current picture" ) );
        connect( changePictureAction_, SIGNAL(triggered()), SLOT(_changePicture()) );

        addAction( closeAction_ = new QAction( IconEngine::get( ICONS::DIALOG_CANCEL ), tr( "Hide" ), this ) );
        closeAction_->setToolTip( tr( "Hide this picture\n  (use the picture list window to make it reappear)" ) );
        closeAction_->setShortcut( QKeySequence::Close );
        connect( closeAction_, SIGNAL(triggered()), SLOT(_hide()) );

        connect( &_staysOnTopAction(), SIGNAL(toggled(bool)), Singleton::get().application(), SIGNAL(configurationChanged()) );
        connect( &_lockPositionAction(), SIGNAL(toggled(bool)), Singleton::get().application(), SIGNAL(configurationChanged()) );
        connect( &_hideFromTaskBarAction(), SIGNAL(toggled(bool)), Singleton::get().application(), SIGNAL(configurationChanged()) );
        connect( &_stickyAction(), SIGNAL(toggled(bool)), Singleton::get().application(), SIGNAL(configurationChanged()) );
        connect( &_roundCornersAction(), SIGNAL(toggled(bool)), Singleton::get().application(), SIGNAL(configurationChanged()) );
        connect( &_useSvgAction(), SIGNAL(toggled(bool)), Singleton::get().application(), SIGNAL(configurationChanged()) );

    }

    //______________________________________________________________________
    void PictureWidget::_paintFilename( QPaintDevice& device, const QRect& clip, bool shadow )
    {

        QPainter painter( &device );
        painter.setClipRect( clip );
        painter.setRenderHints(QPainter::Antialiasing );

        if( shadow ) painter.translate( shadowOffset(), shadowOffset() );

        // create rect for painting text
        QRect rect( PictureWidget::rect() );
        rect.setTop( rect.bottom() - QFontMetrics( font() ).height() - _margins().top() );
        rect.setBottom( rect.bottom() - _margins().bottom() );
        rect.setLeft( _margins().left() );

        painter.setPen(  shadow ? shadowColor() : foregroundColor() );
        painter.drawText( rect, Qt::AlignCenter,  File( attributes().filename() ).localName() );
        painter.end();

    }

    //______________________________________________________________________
    void PictureWidget::_paintEmpty( QPaintDevice& device, const QRect& clip, bool shadow )
    {

        QPainter painter( &device );
        painter.setClipRect( clip );
        painter.setRenderHints(QPainter::Antialiasing );

        // create rect for painting text
        QRect rect( _margins().adjustedRect( this->rect() ) );
        QString text( tr( "Drop picture here or select filename from menu." ) );

        painter.setPen(  shadow ? shadowColor() : foregroundColor() );
        painter.drawText( rect, Qt::AlignCenter|Qt::TextWordWrap,  text );
        painter.end();

    }

}
